import { PolymerElement, html } from '../../../hip/library/polymer/polymer.js';

import './views/ai-builder-view-review-summary.js';
import './views/ai-builder-view-pick-sources.js';
import './views/ai-builder-view-write-report.js';
import './views/ai-builder-view-add-filters.js';
import './views/ai-builder-query-filter.js';
import './views/ai-builder-query-join.js';

import ifn from '../../../pip/library/pi-ifn/pi-ifn.js';
import f1 from '../../../pip/library/pi-f1/pi-f1.js';

class QueryBuilderModel {
  constructor(ns, $, builderQueryEl) {
    this.builderQueryEl = builderQueryEl;
    this.ns = ns;
    this.$ = $;

    this.query = {};

    this.indicies = {
      filter: 0,
      join: 0
    };
  }

  newIndex(name) {
    return (this.indicies[name] += 1);
  }

  updateDefinitions() {
    this.createTableDef();
    this.createFieldsDef();
    this.updateQueryTables();
    this.createTableFields();
    this.updateTableDef();
    this.updateFieldsDef();
  }

  createTableDef() {
    const databaseBucket = HIP.db.gid.queryOne('database', this.ns);
    if (!databaseBucket || !databaseBucket.output) {
      return [];
    }

    const tableDef = databaseBucket.output
      .map((table) => {
        return {
          value: table.name,
          title: table.name
        };
      })
      .sort((a, b) => HIP.data.sortComparator(a, b, 'value', 'asc'));

    this.tableDef = tableDef;
    return tableDef;
  }

  createFieldsDef() {
    const databaseBucket = HIP.db.gid.queryOne('database', this.ns);
    if (!databaseBucket || !databaseBucket.output) {
      return [];
    }

    const fieldsDef = {};
    databaseBucket.output.forEach((table) => {
      fieldsDef[table.name] = table.metadata.fields.map((field) => {
        return {
          value: `${table.name}.${field.name}`,
          title: `${table.name}.${field.name}`
        };
      });
    });

    this.fieldsDef = fieldsDef;
    return fieldsDef;
  }

  createTableFields() {
    const databaseBucket = HIP.db.gid.queryOne('database', this.ns);
    if (!databaseBucket || !databaseBucket.output) {
      return [];
    }

    const tableFields = {};
    databaseBucket.output.forEach((table) => {
      if ((ifn.get('query.tables', this) || []).indexOf(table.name) >= 0) {
        tableFields[table.name] = {
          title: table.name,
          fields: table.metadata.fields.map((field) => {
            return {
              value: `${table.name}.${field.name}`,
              title: field.name
            };
          })
        };
      }
    });

    this.tableFields = tableFields;
    return tableFields;
  }

  createAvailableFields(_uniqueTables = this.uniqueJoinTables()) {
    const fieldsDef = this.createFieldsDef();
    const uniqueTables = !ifn.is(Array, _uniqueTables) ? Object.values(_uniqueTables) : _uniqueTables;

    const availableFields = ifn.reduce(
      (acc, table) => {
        return [...acc, ...table];
      },
      [],
      ifn.filter((table, tableName) => {
        return uniqueTables.indexOf(tableName) >= 0;
      }, fieldsDef)
    );

    this.availableFields = availableFields;
    return availableFields;
  }

  createFilterFields() {}

  resetQuery() {
    this.query = {
      base_table: undefined,
      joins: [],
      tables: [],
      report: {
        fields: []
      },
      limit: {
        filters: [],
        algebra: ''
      }
    };

    this.reconcilePermissions();
    this.disseminateQuery();
  }

  get wasLoaded() {
    return this.__wasLoaded;
  }

  loadQuery(query) {
    if (!f1.is.empty(query)) {
      this.$.view_selector__tab_group.value = 'add_filters';
      this.$.view_selector__tab_group.value = 'pick_sources';

      this.readQuery(JSON.parse(JSON.stringify(query)));
      setTimeout(() => {
        this.$.view_selector__tab_group.value = 'add_filters';
        this.$.view_selector__tab_group.value = 'pick_sources';

        this.readQuery(JSON.parse(JSON.stringify(query)));
        setTimeout(() => {
          this.$.view_selector__tab_group.value = 'add_filters';
          this.$.view_selector__tab_group.value = 'pick_sources';

          this.readQuery(JSON.parse(JSON.stringify(query)));
        }, 1000);
      }, 1000);
    }
  }

  readQuery(query) {
    this.__wasLoaded = true;
    this.query = Object.assign({}, query);

    this.updateDefinitions();
    this.disseminateQuery();
    this.reconcilePermissions();
    this.builderQueryEl.activateView();
  }

  writeQuery() {
    return this.query || {};
  }

  disseminateQuery() {
    this.$.view_review_summary.model = {};
    this.$.view_pick_sources.model = {};
    this.$.view_write_report.model = {};
    this.$.view_add_filters.model = {};

    this.$.view_review_summary.set('model', this);
    this.$.view_pick_sources.set('model', this);
    this.$.view_write_report.set('model', this);
    this.$.view_add_filters.set('model', this);
  }

  reconcilePermissions() {
    const baseTableEl = HIP.db.gid.queryOne('_base_table__select', this.ns);
    if (baseTableEl.value != null || this.query.base_table != null) {
      this.$.view_selector__write_report.locked = false;
    }

    const reportFieldsEl = HIP.db.gid.queryOne('_report_fields__checkbox_group', this.ns);
    const reportFields = reportFieldsEl.values || ifn.get('report.fields', this.query) || [];
    if (reportFields) {
      this.$.view_selector__add_filters.locked = reportFields.length <= 0;
    }
  }

  joinTables() {
    return ifn.filter((o) => o, [...(this.query.joins || [])]);
  }

  uniqueJoinTables() {
    const baseTableSelectEl = HIP.db.gid.queryOne('_base_table__select', this.ns);
    const tablesDict = {};

    if (baseTableSelectEl.value != null) {
      tablesDict[baseTableSelectEl.value] = baseTableSelectEl.value;
    }

    this.joinTables().forEach((join) => {
      if (join.right_table != null) {
        tablesDict[join.right_table] = join.right_table;
      }
    });

    return tablesDict;
  }

  updateQueryTables() {
    this.query.tables = Object.keys(this.uniqueJoinTables());
  }

  updateTableDef() {
    const tableDef = this.tableDef;
    const tableFields = this.tableFields;
    if (!tableDef) {
      return;
    }

    this.$.view_pick_sources.tableDef = tableDef;

    const uniqueSelectedTables = (this.query.tables = Object.keys(this.uniqueJoinTables()));
    const limitedTableDef = tableDef.filter((table) => {
      return uniqueSelectedTables.indexOf(table.value) >= 0;
    });

    const buildQueryFilters = HIP.db.gid.queryBy('_source__build_query_filter', this.ns);
    if (buildQueryFilters) {
      buildQueryFilters.forEach((buildQueryFilter) => {
        buildQueryFilter.tableDef = limitedTableDef;
      });
    }

    const limitedTableFields = ifn.filter((table, key) => {
      return uniqueSelectedTables.indexOf(key) >= 0;
    }, tableFields);

    HIP.db.gid.queryOne('_report_fields__repeat', this.ns).input = limitedTableFields;
    return limitedTableFields;
  }

  updateFieldsDef() {
    const fieldsDef = this.fieldsDef;

    const buildQueryJoins = HIP.db.gid.queryBy('_source__build_query_join', this.ns);
    if (buildQueryJoins) {
      buildQueryJoins.forEach((buildQueryJoin) => {
        buildQueryJoin.fieldsDef = fieldsDef;
      });
    }

    const buildQueryFilters = HIP.db.gid.queryBy('_source__build_query_filter', this.ns);
    if (buildQueryFilters) {
      buildQueryFilters.forEach((buildQueryFilter) => {
        buildQueryFilter.fieldsDef = fieldsDef;
      });
    }
  }

  refreshDropdowns() {
    setTimeout(() => {
      ifn.each((el) => {
        el.refresh();
      }, HIP.browser.queryAll('hi-dropdown', this.ns, this));
    });
  }
}

class PrefabBuilderQueryElement extends HIP.BehaviorPrefab(HIP.BehaviorElement(PolymerElement)) {
  static get is() {
    return 'ai-prefab-builder-query';
  }

  static get template() {
    return html`
      <ai-screen id="_screen" gid="_screen" state="pick_sources">
        <ai-state class="mtm bdr-solid baz bbs clr-gray_40-bdr">
          <hi-step-group id="view_selector__tab_group">
            <hi-step id="view_selector__pick_sources" value="pick_sources" selected> Pick Sources </hi-step>
            <hi-step id="view_selector__write_report" value="write_report" locked> Write Report </hi-step>
            <hi-step id="view_selector__add_filters" value="add_filters" locked> Add Filters </hi-step>
            <hi-step id="view_selector__review_summary" value="review_summary"> Review Summary </hi-step>
          </hi-step-group>
        </ai-state>
        <ai-state states="pick_sources">
          <ai-builder-view-pick-sources id="view_pick_sources" model="[[ model ]]" context="[[ context ]]">
          </ai-builder-view-pick-sources>
        </ai-state>
        <ai-state states="write_report">
          <ai-builder-view-write-report id="view_write_report" model="[[ model ]]" context="[[ context ]]">
          </ai-builder-view-write-report>
        </ai-state>
        <ai-state states="add_filters">
          <ai-builder-view-add-filters id="view_add_filters" model="[[ model ]]" context="[[ context ]]">
          </ai-builder-view-add-filters>
        </ai-state>
        <ai-state states="review_summary">
          <ai-builder-view-review-summary id="view_review_summary" model="[[ model ]]" context="[[ context ]]">
          </ai-builder-view-review-summary>
        </ai-state>
        <ai-state kind="logic" gid="_external_interface">
          <hi-binding source="view_selector__tab_group" signal="value-changed" sink="_screen"> </hi-binding>
        </ai-state>
      </ai-screen>
    `;
  }

  get defaultScript() {
    return {
      system: {
        title: 'Query Builder'
      },
      action: {
        create_query: {},
        update_query: {}
      },
      bucket: {
        context: {
          $reactive_values: {
            text__save_query__button: 'Save Query'
          }
        },
        database: {},
        query: {},
        delimiters: {},
        comparators: {}
      }
    };
  }

  constructor() {
    super();
    this.__updateStateBound = this.__updateState.bind(this);
    this.__loadQueryBound = this.__loadQuery.bind(this);
    this.__screen__onStateEndedBound = this.__screen__onStateEnded.bind(this);
  }

  includeBehaviors(script) {
    this.__behaviors = ifn.get('behavior', script);
    return script;
  }

  connectedCallback() {
    super.connectedCallback();
    this.model = new QueryBuilderModel(this.ns, this.$, this);
    this.model.resetQuery();

    this.$.view_selector__tab_group.addEventListener('selected-index-changed', () => {
      this.activateView();
      this.model.reconcilePermissions();
    });

    this.$.view_selector__tab_group.addEventListener('value-changed', () => {
      this.model.updateDefinitions();
    });

    this.$._screen.addEventListener('state-ended', this.__screen__onStateEndedBound);
  }

  disconnectedCallback() {
    super.disconnectedCallback();

    this.$._screen.removeEventListener('state-ended', this.__updateStateBound);

    const databaseBucket = HIP.db.gid.queryOne('database', this.ns);
    if (databaseBucket) {
      databaseBucket.removeEventListener('invoke-ended', this.__updateStateBound);
    }

    const queryBucket = HIP.db.gid.queryOne('query', this.ns);
    if (queryBucket) {
      queryBucket.removeEventListener('invoke-ended', this.__loadQueryBound);
    }
  }

  _connectedViewCallback() {
    const databaseBucket = HIP.db.gid.queryOne('database', this.ns);
    if (databaseBucket) {
      databaseBucket.addEventListener('invoke-ended', this.__updateStateBound);
    }

    const queryBucket = HIP.db.gid.queryOne('query', this.ns);
    if (queryBucket) {
      queryBucket.addEventListener('invoke-ended', this.__loadQueryBound);
    }

    this.__updateState();
  }

  activateView() {
    const selectedTabEl = this.$.view_selector__tab_group;
    const tabEl = selectedTabEl.children[this.$.view_selector__tab_group.selectedIndex];
    if (!tabEl) {
      return;
    }

    const viewEl = this.$[`view_${tabEl.value}`];

    if (this.__viewEl && this.__viewEl.deactivate) {
      this.__viewEl.deactivate(this.model);
    }

    if (viewEl && viewEl.activate) {
      setTimeout(() => {
        viewEl.activate(this.model);
      });
    }

    this.model.refreshDropdowns();

    this.__viewEl = viewEl;
  }

  __screen__onStateEnded(e) {
    const stateViewEl = this.$[`view_${e.detail.activeState}`];
    if (stateViewEl && typeof stateViewEl.onStateChanged === 'function') {
      stateViewEl.onStateChanged();
    }
  }

  __loadQuery(event) {
    const exportObject = ifn.get('detail.value', event || {});
    if (!exportObject) {
      return;
    }

    const query = ifn.get('query_builder_json', exportObject) || {};
    this.model.loadQuery(query);

    const summary = exportObject;
    const summaryForm = HIP.db.gid.queryOne('_query_summary', this.ns);

    summaryForm.input = summaryForm ? summary : {};
  }

  __updateState() {
    HIP.Polymer.debounce(this, '__updateStateDebouncer', () => {
      this.model.updateDefinitions();
    });
  }
}

HIP.factory.element(PrefabBuilderQueryElement);
