import { PolymerElement, html } from '../../../hip/library/polymer/polymer.js';
import f1 from '../../../pip/library/pi-f1/pi-f1.js';
import ifn from '../../../pip/library/pi-ifn/pi-ifn.js';
/* eslint-disable lit-a11y/anchor-is-valid */

/**
 * @customElement
 * @polymer
 * @appliesMixin BehaviorPrefab
 * @appliesMixin BehaviorElement
 */
class PrefabRecordEditorElement extends HIP.BehaviorPrefab(HIP.BehaviorElement(PolymerElement)) {
  static get is() {
    return 'ai-prefab-record-editor';
  }

  static get template() {
    return html`
      <ai-screen gid="_screen" l-kind="hort">
        <ai-state gid="_state_main" states="main" l-size="12" l-hug="left" class="display-block">
          <section>
            <a href$="[[ context.related_link ]]" download target="_blank"> [[ context.related_title ]] </a>
          </section>

          <section>
            <hi-form id="_record_form" gid="_record_form" aesthetic="invisible" endpoint="api" disable-http>
              <ai-builder-form gid="_record_form__builder"> </ai-builder-form>
              <hi-button gid="_record_form__cancel_button" aesthetic="secondary" slot="controls">
                [[ context.record_form__cancel_button ]]
              </hi-button>
              <hi-button gid="_record_form__submit_button" slot="controls" type="submit">
                [[ context.record_form__submit_button ]]
              </hi-button>
            </hi-form>
          </section>

          <hi-binding source="request_submit_record" sink="_record_form__submit_button" field="trigger"> </hi-binding>
          <hi-binding
            source="_record_form__cancel_button"
            signal="trigger-ended"
            sink="confirm_cancel"
            field="invoke"
            pull-source="configuration"
            pull-field="confirm_cancel_message"
          >
          </hi-binding>

          <hi-binding source="record_type" sink="_record_form__builder"> </hi-binding>
          <hi-binding source="record" sink="_record_form"> </hi-binding>
          <!-- force render -->
          <!-- this is excessive, but if everything is not available we do not render -->
          <hi-binding source="record_type" sink="_record_form" field="flush" stop-value> </hi-binding>
          <!-- force render -->

          <hi-binding source="record_type" sink="_record_form__builder" field="groupType">
            <!-- just stuck this here, not sold on it tho -->
            <di-fn-get path="groupType"></di-fn-get>
          </hi-binding>
        </ai-state>
        <ai-state kind="logic" gid="_external_interface"> </ai-state>
      </ai-screen>
    `;
  }

  get defaultScript() {
    return {
      /**
       * TODO the mode in system does not affect anything!
       * we need it to be a variable, so we can turn stuff on and off
       * the behavior-prefab should wrap it and allow us to respond
       * to changes in both!
       */
      system: {
        state: 'main'
      },
      constants: {
        '@view__submit_button_label': 'Save',
        '@view__cancel_button_label': 'Cancel'
      },
      configuration: {
        confirm_cancel_message: {
          status: 'warning',
          title: 'Are you sure?',
          details: 'This action will discard all the changes you made, press Yes to proceed.',
          controls: [
            {
              type: 'yes',
              name: 'Yes'
            },
            {
              type: 'no',
              name: 'No'
            }
          ]
        }
      },
      life_cycle: {
        initialize: {
          $elements: {
            _record_form__cancel_button: {
              hidden: true
            }
          }
        }
      },
      action: {
        /**
         * request the interface trigger events to invoke submit_record
         */
        request_submit_record: {},
        /**
         * the record is submitted through the form
         */
        submit_record: {},
        /**
         * the action invokes on yes on cancel
         */
        cancel_record: {},
        /**
         * cancel button confirmation message
         */
        confirm_cancel: {
          $message: {
            $is: 'prompt',
            $region: 'view_header',
            $strategy: {
              value: {
                yes: '{ cancel_record }'
              }
            },
            $output: '{ confirm_cancel_proxy }'
          }
        }
      },
      bucket: {
        context: {
          $reactive_values: {
            record_form__submit_button: '@view__submit_button_label',
            record_form__cancel_button: '@view__cancel_button_label'
          }
        },
        record: {}
      },
      io: {
        confirm_cancel_proxy: {
          $is: 'proxy'
        }
      }
    };
  }

  includeBehaviors(script) {
    const restIOName = ifn.first(ifn.get('behavior.rest', script) || []);
    const restIO = ifn.get(`io.${restIOName}`, script);
    const systemMode = ifn.get(`system.mode`, script);

    script = this.__correctFormToSubmissionAction(script, systemMode);
    script = ifn.mergeWith(
      'clobber',
      {
        constants: {
          '@view__record_method': f1.capitalize(systemMode)
        }
      },
      script
    );

    const restConfigure = ifn.mergeWith(
      'clobber',
      {
        return_on_completion: true,
        uri_scheme: 'single-record',
        read_to_bucket: 'record',
        record_base_uri: 'auto',
        api_response_id: 'id'
      },
      ifn.get(`$configure.rest`, restIO || {}) || {}
    );

    return restIOName
      ? this.__supportBehaviorREST(script, restIOName, systemMode, restIO, restConfigure)
      : systemMode === 'read'
      ? this.__readIncludeBehavior(script, restIO, restIOName, restConfigure)
      : script;
  }

  /**
   * connect the form to the submission action
   * - read, we update from the button because the form is in read_only mode and is always invalid
   * - (not read), we connect the form directly to the action so if the validation fails it will not submit
   */
  __correctFormToSubmissionAction(script, systemMode) {
    const submissionToAction =
      systemMode === 'read'
        ? `{ _record_form__submit_button } =>> { submit_record__dead_end } <=< { _record_form }`
        : `{ _record_form } =>> { submit_record }`;

    return ifn.mergeWith('append', script, {
      life_cycle: {
        initialize: {
          $bindings: submissionToAction
        }
      }
    });
  }

  __supportBehaviorREST(script, restIOName, systemMode, restIO, restConfigure) {
    const behaviorMethod = `__${systemMode}IncludeBehavior`;
    if (!this[behaviorMethod]) {
      return script;
    }

    script = this.__linkIncludeBehavior(script, restIO, restIOName, restConfigure);
    script = this[behaviorMethod](script, restIO, restIOName, restConfigure);
    return script;
  }

  __postIncludeBehavior(script, restIO, restIOName, restConfigure) {
    return this.__wrapInErrorHandler(
      this.__notificationIncludeBehaviorAddendum(
        ifn.mergeWith('append', script, {
          io: {
            [restIOName]: {
              uri: `${restIO.uri}@trailing_slash`,
              $input: '{ submit_record }',
              method: 'post'
            }
          }
        }),
        restIOName,
        restConfigure
      ),
      restIOName,
      restConfigure
    );
  }

  __putIncludeBehavior(script, restIO, restIOName, restConfigure) {
    return this.__wrapInErrorHandler(
      this.__notificationIncludeBehaviorAddendum(
        ifn.mergeWith('append', script, {
          io: {
            [restIOName]: {
              uri: `${restIO.uri}@trailing_slash`,
              $input: '{ submit_record }',
              method: 'put'
            }
          }
        }),
        restIOName,
        restConfigure
      ),
      restIOName,
      restConfigure
    );
  }

  __createIncludeBehavior(script, restIO, restIOName, restConfigure) {
    return this.__wrapInErrorHandler(
      this.__notificationIncludeBehaviorAddendum(
        ifn.mergeWith('append', script, {
          io: {
            [restIOName]: {
              uri: `${restIO.uri}@trailing_slash`,
              $input: '{ submit_record }',
              method: 'post'
            }
          }
        }),
        restIOName,
        restConfigure
      ),
      restIOName,
      restConfigure
    );
  }

  __readIncludeBehavior(script, restIO, restIOName, restConfigure) {
    const readBehavior = restIOName
      ? this.__readIncludeBehaviorAddendum(script, restIO, restIOName, restConfigure)
      : {};

    return ifn.mergeWith('append', script, readBehavior, {
      configuration: {
        record_editor__read_mode__no_submit: {
          status: 'warning',
          heading: 'Submission Disabled',
          text: 'You cannot submit a form in read-only mode.',
          controls: [
            {
              type: 'yes',
              name: 'Ok'
            }
          ]
        },
        record_editor__read_mode__merge_over: {
          readonly: true
        }
      },
      life_cycle: {
        initialize: {
          $elements: {
            _record_form: {
              mode: 'display'
            },
            _record_form__builder: {
              mergeOver: '{ configuration.*.record_editor__read_mode__merge_over }'
            },
            _record_form__cancel_button: {
              hidden: true
            }
          }
        }
      },
      action: {
        submit_record__dead_end: {
          $do: ['=>> { hip_fn.prompt } <=< { configuration.*.record_editor__read_mode__no_submit }']
        }
      }
    });
  }

  __updateIncludeBehavior(script, restIO, restIOName, restConfigure) {
    return this.__wrapInErrorHandler(
      this.__notificationIncludeBehaviorAddendum(
        ifn.merge(this.__readIncludeBehaviorAddendum(script, restIO, restIOName, restConfigure), {
          io: {
            [restIOName]: {
              uri: `${restIO.uri}/${this.__getRecordKey(restIO)}@trailing_slash`,
              $input: '{ submit_record }',
              method: 'put'
            }
          }
        }),
        restIOName,
        restConfigure
      ),
      restIOName,
      restConfigure
    );
  }

  __reviewIncludeBehavior(script, restIO, restIOName, restConfigure) {
    return this.__wrapInErrorHandler(
      this.__notificationIncludeBehaviorAddendum(
        ifn.merge(this.__readIncludeBehaviorAddendum(script, restIO, restIOName, restConfigure), {
          io: {
            [restIOName]: {
              uri: `${restIO.uri}/${this.__getRecordKey(restIO)}@trailing_slash`,
              $input: '{ submit_record }',
              manual: true,
              method: 'put'
            }
          }
        }),
        restIOName,
        restConfigure
      ),
      restIOName,
      restConfigure
    );
  }

  __deleteIncludeBehavior(script, restIO, restIOName, restConfigure) {
    return this.__wrapInErrorHandler(
      this.__notificationIncludeBehaviorAddendum(
        ifn.merge(this.__readIncludeBehaviorAddendum(script, restIO, restIOName, restConfigure), {
          constants: {
            '@view__submit_button_label': 'Delete'
          },
          io: {
            [restIOName]: {
              uri: `${restIO.uri}/${this.__getRecordKey(restIO)}@trailing_slash`,
              $input: '{ submit_record }',
              method: 'delete'
            }
          },
          life_cycle: {
            initialize: {
              $elements: {
                _record_form__submit_button: {
                  theme: 'red'
                }
              }
            }
          }
        }),
        restIOName,
        restConfigure
      ),
      restIOName,
      restConfigure
    );
  }

  __editIncludeBehavior(script, restIO, restIOName, restConfigure) {
    return this.__wrapInErrorHandler(
      this.__notificationIncludeBehaviorAddendum(
        ifn.merge(this.__readIncludeBehaviorAddendum(script, restIO, restIOName, restConfigure), {
          io: {
            [restIOName]: {
              uri: `${restIO.uri}/${this.__getRecordKey(restIO)}@trailing_slash`,
              $input: '{ submit_record }',
              method: 'patch'
            }
          }
        }),
        restIOName,
        restConfigure
      ),
      restIOName,
      restConfigure
    );
  }

  /**
   * get the record key to lookup the record on
   * @param {*} restIO
   */
  __getRecordKey(restIO) {
    // @deprecated
    const recordId =
      ifn.get('$configure.rest.record_id', restIO) ||
      ifn.get('$configure.rest.recordPath', restIO) ||
      ifn.get('$configure.rest.recordId', restIO) ||
      ifn.get('$additional.recordId', restIO) ||
      ifn.get('$additional.key', restIO);

    if (recordId != null) {
      const reducedRecordId = recordId.replace('((', '').replace('))', '');

      return `(( ${reducedRecordId} ))`;
    }

    // something like `{ :id }` instead of `location.parameters.id`
    return ifn.get('$configure.rest.record_ref', restIO);
  }

  __linkIncludeBehavior(script, restIO, restIOName, restConfigure) {
    if (restConfigure.return_on_completion === false) {
      return script;
    }

    const systemMode = ifn.get(`system.mode`, script);
    const baseURI =
      restConfigure.record_base_uri === 'auto'
        ? ifn.first(location.hash.split('/editor') || [])
        : restConfigure.record_base_uri;

    const goToManagerURI =
      restConfigure.go_to_manager_uri != null ? restConfigure.go_to_manager_uri : `${baseURI}/manager/`;

    const goToEditorURI =
      restConfigure.go_to_editor_uri != null
        ? restConfigure.go_to_editor_uri
        : `{ location[nth_3] }/{ ${restIOName}.*.${restConfigure.api_response_id} }/editor/review`;

    const goToViewerURI =
      restConfigure.go_to_viewer_uri != null
        ? restConfigure.go_to_viewer_uri
        : `${baseURI}/viewer/read/${this.__getRecordKey(restIO)}`;

    const goToOnSuccessURI =
      restConfigure.go_to_on_success != null
        ? restConfigure.go_to_on_success
        : restConfigure.uri_scheme === 'parent-child'
        ? systemMode === 'create'
          ? goToEditorURI
          : goToManagerURI
        : restConfigure.go_to_manager_uri != null
        ? restConfigure.go_to_manager_uri
        : `${baseURI}/manager/`;

    const goToOnErrorURI = restConfigure.go_to_on_error != null ? restConfigure.go_to_on_error : `${baseURI}/manager/`;

    return ifn.mergeWith('append', script, {
      action: {
        go_to_on_success: {
          $do: `=>> { location.goto } <-< ${goToOnSuccessURI}@trailing_slash`
        },
        go_to_on_error: {
          $do: `=>> { location.goto } <-< ${goToOnErrorURI}@trailing_slash`
        },
        go_to_manager: {
          $do: `=>> { location.goto } <-< ${goToManagerURI}@trailing_slash`
        },
        go_to_editor: {
          $do: `=>> { location.goto } <-< ${goToEditorURI}@trailing_slash`
        },
        go_to_viewer: {
          $do: `=>> { location.goto } <-< ${goToViewerURI}@trailing_slash`
        }
      }
    });
  }

  __wrapInErrorHandler(script, restIOName) {
    return ifn.mergeWith('append', script, {
      io: {
        [restIOName]: {
          allowFail: true,
          $ahp: {
            network: {
              intercept_errors: false
            }
          },
          $error: ['=>> prefab:informOnError']
        }
      }
    });
  }

  informOnError(networkResponse) {
    const errors = ifn.get('error.metadata.errors', networkResponse);
    const message = ifn.get('error.metadata.message', networkResponse);

    if (!f1.is.empty(errors)) {
      this.$._record_form.validateCustomValidity(errors);
    } else if (ifn.is(String, message)) {
      HIP.fn.notify({ status: 'error', title: message });
    }
  }

  /**
   *
   * @param {*} script
   * @param {*} humanIOName what is the name of the operation (eg. create)
   * @param {*} restIOName what is the io involved in the operation
   */
  __notificationIncludeBehaviorAddendum(script, restIOName, restConfigure) {
    if (!restConfigure.return_on_completion) {
      return script;
    }

    return ifn.mergeWith('append', script, {
      io: {
        [restIOName]: {
          $finish: `{ go_to_on_success }`,
          $error: `{ go_to_on_error }`
        }
      }
    });
  }

  __readIncludeBehaviorAddendum(script, restIO, restIOName, restConfigure) {
    return ifn.mergeWith('append', script, {
      io: {
        [`read_${restIOName}`]: {
          uri: `${restIO.uri}/${this.__getRecordKey(restIO)}@trailing_slash`,
          $output: `(( ${restConfigure.read_to_bucket} ))`,
          $startup: {
            $state: 'main'
          }
        }
      }
    });
  }

  _alterContext(context) {
    context = super._alterContext(context);

    if (!context.related_link) {
      context.related_title = '';
    }

    return context;
  }
}

HIP.factory.element(PrefabRecordEditorElement);
