import { afterNextRender } from '../../../hip/library/polymer/polymer.js';
import { PolymerElement, html } from '../../../hip/library/polymer/polymer.js';
import ifn from '../../../pip/library/pi-ifn/pi-ifn.js';

/**
 * @customElement
 * @polymer
 * @appliesMixin BehaviorPrefab
 * @appliesMixin BehaviorElement
 */
class PrefabViewsWorkflowElement extends HIP.BehaviorPrefab(HIP.BehaviorElement(PolymerElement)) {
  static get is() {
    return 'ai-prefab-views-workflow';
  }

  static get template() {
    return html`
      <ai-screen gid="_screen" l-kind="hort">
        <ai-state id="_state_steps" gid="_state_steps" states="main,no_step" l-size="3">
          <hi-step-group
            id="_steps__step_group"
            gid="_steps__step_group"
            manual-furthest-index
            on-selected-ended="__stepGroup__onSelectedEnded"
            aesthetic="stacked"
            kind="navigation"
          >
            <hi-repeat id="_steps__step_group__repeat" gid="_steps__step_group__repeat" as="step">
              <template>
                <hi-step gid="_step" value$="[[ step.name ]]" kind$="[[ _getStepKind(step) ]]" aesthetic="stacked">
                  <span hidden$="[[ _showTitle(step) ]]"> [[ step.title ]] </span>
                  <hi-link href$="[[ _getStepGoToURI(step) ]]" hidden$="[[ !_showTitle(step) ]]">
                    [[ step.title ]]
                  </hi-link>
                </hi-step>
              </template>
            </hi-repeat>
          </hi-step-group>

          <hi-binding source="region_group" sink="_steps__step_group__repeat"> </hi-binding>

          <hi-binding source="_steps__step_group" signal="selected-ended" sink="activated_view">
            <di-fn-get path="detail.value"></di-fn-get>
            <di-fn-init hoist-path="view_name"></di-fn-init>
          </hi-binding>
        </ai-state>

        <ai-state id="_state_step" gid="_state_step" states="main" l-kind="vert">
          <ai-block id="_sub_view_region" gid="_sub_view_region" region="sub_view"></ai-block>
        </ai-state>

        <ai-state id="_state_pending" gid="_state_pending" states="no_step" l-kind="vert">
          <h2>No Step Active</h2>
        </ai-state>
      </ai-screen>
    `;
  }

  get properties() {
    return {
      activeStepIndex: {
        type: Number
      }
    };
  }

  get defaultScript() {
    return {
      system: {
        state: 'main'
      },
      configuration: {
        initial_step_name: ''
      },
      action: {
        /** @deprecated listen for activated_view */
        changed_step: {
          $input: '{ activated_view }'
        },
        activated_view: {}
      },
      bucket: {
        steps: {},
        active_view: {
          $do: ['=>> { _steps__step_group.value } <=< { active_view.*.view_name }']
        }
      }
    };
  }

  constructor(...args) {
    super(...args);
    this.__instances = {};
  }

  disconnectedCallback() {
    super.disconnectedCallback();

    this.__instances.regionGroup.removeEventListener('output-changed', this.__regionGroup__onOutputChangedBound);
  }

  _connectedViewCallback() {
    const ns = this.ns;

    this.__instances.configuration = HIP.db.gid.queryOne('configuration', ns);
    this.__instances.regionGroup = HIP.db.gid.queryOne('region_group', ns);

    this.__regionGroup__onOutputChangedBound = this.__regionGroup__onOutputChanged.bind(this);
    this.__instances.regionGroup.addEventListener('output-changed', this.__regionGroup__onOutputChangedBound);

    afterNextRender(this, () => {
      afterNextRender(this, () => {
        this._updateCurrentStep();
      });
    });
  }

  includeBehaviors(language) {
    return this._setPrefabMode(language);
  }

  _setPrefabMode(language) {
    const prefabMode = ifn.get('system.mode', language);

    // accept the region from a bucket and not hard-coded
    //
    if (prefabMode === 'dynamic') {
      language = ifn.merge(language, {
        bucket: {
          steps: {
            $output: ['{ region_group }']
          }
        }
      });
    }

    return language;
  }

  /**
   * go to the next step in the sequence
   */
  nextStep() {
    return this.goToStepIndex(this.activeStepIndex + 1);
  }

  /**
   * go to the previous step in the sequence
   */
  previousStep() {
    return this.goToStepIndex(this.activeStepIndex - 1);
  }

  /**
   *
   */
  goToStepIndex(stepIndex) {
    const steps = this.__readActiveSteps();
    if (!steps) {
      return Promise.resolve();
    }

    const activeStep = ifn.first(ifn.filter((step) => step.index === stepIndex, steps));
    if (activeStep) {
      return this.goToStep(activeStep.name);
    } else {
      return Promise.resolve();
    }
  }

  /**
   * go to a permitted step
   */
  goToStep(goToName) {
    const steps = this.__readActiveSteps();
    if (!steps) {
      return Promise.resolve();
    }

    const activeStep = steps[goToName];

    this.activeStep = activeStep;
    this.activeStepIndex = activeStep.index;

    HIP.db.gid.queryOne('active_view', this.ns).invoke({
      view_name: activeStep.name
    });

    return Promise.resolve();
  }

  /**
   * when the step-group has a selection we inform on the step change. This could be from:
   * - link
   * - uri change
   * - programmatic
   */
  __stepGroup__onSelectedEnded(event) {
    const steps = this.__steps;
    const stepEl = event.detail.item;
    const activeStepName = stepEl.value;
    const step = steps[activeStepName];

    this._viewStep(step);
  }

  /**
   * load the actual screen
   * @see __changedStep__onInvokeEnded, these are made to be called in sequence
   */
  _viewStep(step) {
    if (step == null || step.name == null) {
      return;
    }

    const subViewRegionEl = this.$._sub_view_region;
    const regionTarget = step.name;

    this._createSubView({
      subViewRegionEl,
      regionTarget
    });
  }

  __regionGroup__onOutputChanged() {
    this.__readStepsFromRegionGroup();
  }

  /**
   * add a numeric index to each steppable step so we can use indicies with step-group
   * if a route has `$tag` then it is targetting a parent, so it should be counted
   */
  __readActiveSteps() {
    if (this.__steps) {
      return this.__steps;
    }

    return this.__readStepsFromRegionGroup();
  }

  __readStepsFromRegionGroup() {
    let index = 0;

    return (this.__steps = ifn.map((o) => {
      if (o.$tag != null) {
        o.index = index;
        index += 1;
      }

      return o;
    }, ifn.get('output', this.__instances.regionGroup) || {}));
  }

  /**
   * either the uri has:
   *
   * - # so we just pass along
   * - if not then we add the parent hash (relative)
   *
   * @param {*} step
   */
  _getStepGoToURI(step) {
    const goToURI = ifn.get('$configure.workflow.goto_uri', step);
    const parentURIHash = HIP.db.gid.queryOne('location', this.ns).output.parent_hash;

    // undefined will unset the hi-link which effectively makes it disabled
    //
    return goToURI ? (goToURI.indexOf('#') >= 0 ? goToURI : `${parentURIHash}${goToURI}`) : undefined;
  }

  /**
   * $tag means that it is a link and not a title
   */
  _getStepKind(step) {
    return step.$tag == null ? 'title' : 'navigation';
  }

  /**
   * $tag means that it is a link and not a title
   */
  _showTitle(step) {
    return step.$tag == null ? false : true;
  }

  /**
   *
   */
  _onLocationChanged() {
    this._updateCurrentStep();
  }

  _updateCurrentStep() {
    const ns = this.ns;
    const locationEl = HIP.db.gid.queryOne('location', ns);
    const configurationEl = HIP.db.gid.queryOne('configuration', ns);
    const step_name = ifn.get('parameters.view_name', locationEl) || ifn.get('parameters.step_name', locationEl);

    if (this._isReady !== true) {
      configurationEl.invoke({
        ...configurationEl.output,
        initial_step_name: step_name
      });

      return Promise.resolve();
    } else if (step_name != null) {
      return this.goToStep(step_name);
    }
  }
}

HIP.factory.element(PrefabViewsWorkflowElement);
