
import { LitElement, html, css } from '../../../hip/library/lit-element/lit-element.js';
import { AsLitHIC, BehaviorElement, HIP } from '../../../eip/element/hi-component/hi-component.js';
import { stylesTheme } from '../../../eip/element/hi-component/hi-component-theme.js';
import '../../../eip/element/hi-button/hi-close.js';
import '../../../eip/element/hi-icon/hi-icon.js';
export const styles = css`
  :host {
    --hi-alert--color-theme: var(--hi-clr--theme-1);
    --hi-alert--color-theme-text: var(--hi-clr--text);
  }

  :host {
    display: flex;
    position: relative;
    background-color: var(--hi-clr--white);
    color: var(--hi-alert--color-theme-text);

    border-radius: var(--hi-shape--radius-2);
    border: var(--hi-px-1) solid var(--hi-clr--theme-3);
    border-top-width: var(--hi-px-4);

    min-height: var(--hi-px-32);
    min-width: var(--hi-px-32);
    margin: var(--hi-px-16) 0;
  }

  #alertMessage,
  #actionContainer,
  #actionContainer ::slotted([slot]),
  #alertMessage ::slotted([slot]) {
    color: var(--hi-alert--color-theme-text);
  }

  #closeButton {
    position: absolute;
    z-index: 100;
    right: var(--hi-px-8);
    top: var(--hi-px-8);
  }

  :host([mode='presentation']) #closeButton {
    display: none;
  }

  #themeIcon {
    position: absolute;
    left: var(--hi-px-13);
    top: var(--hi-px-13);
  }

  #alertMessage {
    width: 100%;
    margin: var(--hi-px-16) var(--hi-px-36) var(--hi-px-16) var(--hi-px-48);
    font-weight: var(--hi-font--weight-bold);
    font-size: var(--hi-font-size--text);
    line-height: var(--hi-font-lh--medium);
    letter-spacing: 0.02em;
  }

  #actionContainer ::slotted([slot]),
  #alertMessage ::slotted([slot]),
  #alertMessage {
    display: block;
  }

  #alertMessage ::slotted([slot='details']) {
    font-weight: var(--hi-font--weight-light) !important;
    padding: var(--hi-px-4) 0 !important;
    margin-top: 0 !important;
    letter-spacing: 0 !important;
  }

  #actionContainer ::slotted([slot='action']) {
    margin-top: var(--hi-px-4) !important;
    padding: var(--hi-px-8) 0 !important;
    font-weight: var(--hi-font--weight-light) !important;
    letter-spacing: 0 !important;
  }

  :host([remove-ready]) {
    position: fixed;
    top: -999px;
    left: -999px;
  }

  :host([disabled]),
  :host([readonly]) {
    pointer-events: all !important;
  }

  :host([disabled]) #closeButton,
  :host([readonly]) #closeButton {
    cursor: not-allowed;
    pointer-events: none;
  }
`;
export class AlertElement extends BehaviorElement(AsLitHIC(LitElement)) {
  constructor() {
    super(...arguments);
    this.kind = "remove";
  }
  static get is() {
    return "hi-alert";
  }
  static get styles() {
    return [stylesTheme, styles];
  }
  static get properties() {
    return {
      kind: {
        type: String
      },
      mode: {
        type: String
      }
    };
  }
  render() {
    const themeTo = (to) => {
      const icon = to === "icon";
      return {
        info: icon ? "info" : "info alert",
        warning: icon ? "warning" : "warning alert",
        error: icon ? "stop" : "error alert",
        success: icon ? "check_circle" : "success alert",
        failover: icon ? "info" : "primary alert"
      };
    };
    const themeToIcon = themeTo("icon");
    const themeToStatus = themeTo("status");
    const iconKind = themeToIcon[this.theme] ? themeToIcon[this.theme] : "info";
    const alertStatus = themeToStatus[this.theme] || themeToStatus["failover"];
    return html`
      <hi-close
        id="closeButton"
        @trigger-ended=${this.close}
        size="small"
        theme="gray-dark"
        aria-label="Close ${alertStatus}"
        .disabled=${this.disabled}
      >
      </hi-close>
      <hi-icon
        id="themeIcon"
        kind=${iconKind}
        ariaLabel=${alertStatus}
        theme=${this.disabled ? "disabled" : this.theme}
        size="large"
        type="material"
      >
      </hi-icon>
      <div id="alertMessage">
        <slot></slot>
        <slot name="details"></slot>
        <div id="actionContainer">
          <slot name="action"></slot>
        </div>
      </div>
    `;
  }
  static get api() {
    return ["close", "open"];
  }
  static get valueEvent() {
    return "close-ended";
  }
  static get autoBinding() {
    return {
      signal: "close-ended",
      field: "close"
    };
  }
  connectedCallback() {
    super.connectedCallback();
    this.setAttribute("role", "alert");
  }
  alert(packet) {
    const title = packet.title || packet.heading;
    const details = packet.details || packet.description;
    const titleHTML = !title ? "" : `<header><strong>${title}</strong></header>`;
    const descriptionHTML = !details ? "" : `<span slot="details">${details}</span>`;
    this.innerHTML = `${titleHTML}${descriptionHTML}`;
    this.open();
  }
  open() {
    if (!this.disabled) {
      this.removeAttribute("hidden");
    }
  }
  close() {
    if (!this.disabled) {
      switch (this.kind) {
        case "remove":
          this.setAttribute("remove-ready", "");
          this.addEventListener("focusout", () => {
            this.remove();
          });
          break;
        case "hide":
          this.setAttribute("hidden", "hidden");
          break;
        case "passive":
          break;
        default:
          break;
      }
    }
  }
}
export default HIP.factory.element(AlertElement);
