
import { PolymerElement, html } from '../../../hip/library/polymer/polymer.js';
import { Debouncer } from '../../../hip/library/polymer/polymer.js';
import { idlePeriod } from '../../../hip/library/polymer/polymer.js';
import { AsHIC, BehaviorElement, HIP } from '../../../eip/element/hi-component/hi-component.js';
import { BehaviorInput } from '../../../eip/element/hi-behavior-input/hi-behavior-input.js';
import ess from '../../../pip/library/pi-ess/pi-ess.js';
import f1 from '../../../pip/library/pi-f1/pi-f1.js';
export class EditorElement extends BehaviorInput(BehaviorElement(AsHIC(PolymerElement))) {
  constructor() {
    super();
    this.__mutex__internalWrite = false;
    this.value = null;
    this.configuration = this.expandedConfig;
  }
  static get is() {
    return "hi-editor";
  }
  static get template() {
    return html`
      <style is="custom-style" include="hi-theme"></style>
      <style>
        :host {
          display: block;
          /* the editor will be at least 8 rem, so we do not shift the layout as much */
          min-height: 8rem;
        }
      </style>
      <slot></slot>
    `;
  }
  static get properties() {
    return {
      value: {
        type: String,
        reflectToAttribute: true,
        notify: true
      },
      configuration: {
        type: Object
      },
      mode: {
        type: String,
        reflectToAttribute: true
      }
    };
  }
  static get observers() {
    return ["__valueWatcher(value)", "__configurationObserver(configuration, mode)"];
  }
  static get valueEvent() {
    return "value-changed";
  }
  static get autoBinding() {
    return {
      signal: "value-changed",
      field: "value"
    };
  }
  connectedCallback() {
    super.connectedCallback();
    this.disallowEnter = true;
    this._includeGlobalStyles();
    this._createTextarea();
    this.__loadCKEditor();
  }
  disconnectedCallback() {
    super.disconnectedCallback();
    this._destroyEditor();
  }
  _createTextarea() {
    this.el = document.createElement("textarea");
    this.el.id = `${this.id || ""}-editor-id-${f1.id()}`;
    this.el.style.visibility = "hidden";
    this.appendChild(this.el);
  }
  __loadCKEditor() {
    return HIP.util.waitForLibrary("@eab-hip/ckeditor5", "ClassicEditor").then(() => {
      this._destroyEditor();
      this._createEditor();
    }).catch(() => {
      this._destroyEditor();
    });
  }
  _createEditor() {
    if (!window["ClassicEditor"]) {
      return;
    }
    const configuration = this.__getConfigurationByMode(this.configuration, this.mode);
    window["ClassicEditor"].create(this.el, configuration).then((editor) => {
      this.editor = editor;
      editor.setData(f1.format("textarea", {}, this.value || ""));
      editor.model.document.on("change", () => {
        this.__mutex__internalWrite = true;
        this.value = editor.getData() ? editor.getData() : null;
        this.__mutex__internalWrite = false;
      });
      const doc = this.editor.editing.view.document;
      doc.on("selectionChangeDone", (e) => {
        this.dispatchEvent(
          new CustomEvent("text-selection-changed", {
            detail: e
          })
        );
      });
      doc.on("focus", (e) => {
        this.dispatchEvent(
          new CustomEvent("focus-changed", {
            detail: e
          })
        );
      });
      editor.ui.focusTracker.on("change:isFocused", (evt, name, isFocused) => {
        if (!isFocused) {
          this.commit();
        }
      });
      editor.stylesSet.add("default", [
        { name: "CSS Style", element: "ol", attributes: { kind: "data" } },
        { name: "CSS Style", element: "ul", attributes: { kind: "data" } }
      ]);
    }).catch((error) => {
      ess.ele(this, {
        status: "failure",
        action: "create",
        data: error
      });
    });
  }
  get expandedConfig() {
    const colors = [
      { label: "Light Red", color: "var(--hi-clr--red-1)" },
      { label: "Red", color: "var(--hi-clr--red-5)" },
      { label: "Light Orange", color: "var(--hi-clr--orange-1)" },
      { label: "Orange", color: "var(--hi-clr--orange-5)" },
      { label: "Light Yellow", color: "var(--hi-clr--yellow-1)" },
      { label: "Yellow", color: "var(--hi-clr--yellow-5)" },
      { label: "Light Green", color: "var(--hi-clr--green-1)" },
      { label: "Green", color: "var(--hi-clr--green-5)" },
      { label: "Light Teal", color: "var(--hi-clr--teal-1)" },
      { label: "Teal", color: "var(--hi-clr--teal-5)" },
      { label: "Light Blue", color: "var(--hi-clr--blue-1)" },
      { label: "Blue", color: "var(--hi-clr--blue-5)" },
      { label: "Light Cyan", color: "var(--hi-clr--cyan-1)" },
      { label: "Cyan", color: "var(--hi-clr--cyan-5)" },
      { label: "Light Purple", color: "var(--hi-clr--purple-1)" },
      { label: "Purple", color: "var(--hi-clr--purple-5)" },
      { label: "Light Magenta", color: "var(--hi-clr--magenta-1)" },
      { label: "Magenta", color: "var(--hi-clr--magenta-5)" },
      { label: "White", color: "var(--hi-clr--white)", hasBorder: true },
      { label: "Light Gray", color: "var(--hi-clr--gray-2)", hasBorder: true },
      { label: "Gray", color: "var(--hi-clr--gray-5)" },
      { label: "Dark Gray", color: "var(--hi-clr--gray-9)" },
      { label: "Black", color: "var(--hi-clr--black)" }
    ];
    return {
      fontSize: {
        options: [8, 10, 12, "default", 16, 18, 20]
      },
      fontColor: {
        colors
      },
      fontBackgroundColor: {
        colors
      }
    };
  }
  get defaultConfig() {
    return {
      ...this.expandedConfig,
      toolbar: [...this.minimalConfig.toolbar, "|", "alignment", "mergetags"]
    };
  }
  get minimalConfig() {
    return {
      ...this.expandedConfig,
      toolbar: [
        "heading",
        "|",
        "bold",
        "italic",
        "underline",
        "strikethrough",
        "|",
        "bulletedList",
        "numberedList",
        "|",
        "link",
        "blockQuote",
        "|",
        "undo",
        "redo"
      ]
    };
  }
  get validity() {
    return this._computeValidity({
      valueMissing: this.required && this.value == null
    });
  }
  __configurationObserver(configuration, mode) {
    this._debounce = Debouncer.debounce(this._debounce, idlePeriod, () => {
      this._destroyEditor();
      this._createEditor();
    });
  }
  __valueWatcher(value) {
    if (this.editor && !this.__mutex__internalWrite) {
      this.editor.setData(value || "");
    }
  }
  __getConfigurationByMode(configuration, mode) {
    if (mode === "minimal") {
      return this.minimalConfig;
    } else if (mode === "expanded") {
      return this.expandedConfig;
    } else if (mode === "custom" || !mode && configuration && configuration.toolbar) {
      return f1.merge("over", configuration || {}, this.expandedConfig || {});
    } else {
      return this.defaultConfig;
    }
  }
  _destroyEditor() {
    if (this.editor) {
      this.editor.destroy();
      this.editor = void 0;
    }
  }
}
export default HIP.factory.element(EditorElement);
(function setMinDimensions() {
  const style = document.createElement("style");
  style.type = "text/css";
  style.innerHTML = `.ck-editor__editable {
    min-height: 6rem;
    width: 100%;
  }`;
  document.getElementsByTagName("head")[0].appendChild(style);
})();
