
import { html, LitElement } from '../../../hip/library/lit-element/lit-element.js';
import { render, nothing } from '../../../hip/library/lit-html/lit-html.js';
import { AsLitHIC, BehaviorElement } from '../../../eip/element/hi-component/hi-component.js';
import { HIP } from '../../../eip/element/hi-component/hi-component.js';
import dom from '../../../pip/library/pi-dom/pi-dom.js';
import f1 from '../../../pip/library/pi-f1/pi-f1.js';
import './hi-menu.js';
import './hi-sub-menu.js';
import './hi-pick.js';
export class MenuProxyElement extends BehaviorElement(AsLitHIC(LitElement)) {
  constructor() {
    super(...arguments);
    this.opened = false;
    this.childrenAttr = "children";
    this.autoTetheredMenuId = `menuControl__${f1.uuid()}`;
    this.__controlEl__onTriggerEndedBound = this.__controlEl__onTriggerEnded.bind(this);
    this.__menuControl__repeatEventBound = this.__menuControl__repeatEvent.bind(this);
    this.__menuControl__onOpenedChangedBound = this.__menuControl__onOpenedChanged.bind(this);
  }
  static get is() {
    return "hi-menu-proxy";
  }
  static get properties() {
    return {
      opened: {
        type: Boolean
      },
      input: {
        type: Array
      },
      menuSettings: {
        type: Object,
        attribute: "menu-settings"
      },
      forControl: {
        type: String,
        attribute: "for-control"
      },
      forMenu: {
        type: String,
        attribute: "for-menu"
      },
      childrenAttr: {
        type: String,
        attribute: "children-attr"
      }
    };
  }
  static get api() {
    return ["activate", "deactivate"];
  }
  get tetheredControl() {
    return HIP.browser.queryOne(this.forControl, this.ns);
  }
  get tetheredMenu() {
    return HIP.browser.queryOne(this.tetheredMenuId, this.ns);
  }
  get tetheredMenuId() {
    return f1.safe(this.autoTetheredMenuId, this.forMenu);
  }
  disconnectedCallback() {
    super.disconnectedCallback();
    this.deactivate();
  }
  update(props) {
    super.update(props);
    if (props.has("forControl") && this.forControl) {
      this.__unlistenToControl(props["forControl"], this.ns);
      this.__listenToControl(this.forControl, this.ns);
    }
    const ifRelatedSettingChanges = props.has("menuSettings") || props.has("forControl") || props.has("input") || props.has("forMenu");
    if (ifRelatedSettingChanges && !f1.is.empty(this.input) && this.opened) {
      this.deactivate();
      this.activate();
    } else if (this.opened) {
      this.activate();
    } else if (!this.opened) {
      this.deactivate();
    }
  }
  activate() {
    this.opened = true;
    const menuEl = this.__createMenu();
    if (menuEl) {
      menuEl.activate("true", {
        detail: {
          srcElement: this.tetheredControl
        }
      });
    }
  }
  deactivate() {
    const menuEl = this.tetheredMenu;
    if (menuEl) {
      menuEl.deactivate();
    }
    this.__destroyMenu();
    this.opened = false;
  }
  __listenToControl(forControl, ns) {
    const buttonEl = HIP.browser.queryOne(forControl, ns);
    if (buttonEl) {
      buttonEl.addEventListener("trigger-ended", this.__controlEl__onTriggerEndedBound);
    }
  }
  __unlistenToControl(forControl, ns) {
    const buttonEl = HIP.browser.queryOne(forControl, ns);
    if (buttonEl) {
      buttonEl.removeEventListener("trigger-ended", this.__controlEl__onTriggerEndedBound);
    }
  }
  __controlEl__onTriggerEnded() {
    this.opened = true;
  }
  __destroyMenu() {
    const tetheredMenu = this.tetheredMenu;
    if (tetheredMenu) {
      tetheredMenu.removeEventListener("selected-ended", this.__menuControl__repeatEventBound);
      tetheredMenu.removeEventListener("open-ended", this.__menuControl__repeatEventBound);
      tetheredMenu.removeEventListener("opened-changed", this.__menuControl__onOpenedChangedBound);
      render(nothing, tetheredMenu);
    }
    const parentEl = HIP.browser.queryOne(`[menu-parent-proxy-id="${this.autoTetheredMenuId}"]`);
    if (parentEl) {
      dom.destroy(parentEl);
    }
  }
  __createMenu() {
    const menuEl = this.__buildMenu();
    if (!menuEl) {
      return;
    }
    menuEl.addEventListener("selected-ended", this.__menuControl__repeatEventBound);
    menuEl.addEventListener("open-ended", this.__menuControl__repeatEventBound);
    menuEl.addEventListener("opened-changed", this.__menuControl__onOpenedChangedBound);
    dom.setAttrs(menuEl, {
      "menu-proxied": true,
      for: this.forControl
    });
    this.__applyMenuSettings(menuEl);
    render(this.__renderMenuItems(), menuEl);
    menuEl.activate();
    return menuEl;
  }
  __buildMenu() {
    const menuHookEl = HIP.browser.queryOne(this.forMenu, this.ns) || dom.create("body", "div");
    const parentMenuEl = menuHookEl.localName === "hi-menu" ? void 0 : menuHookEl;
    if (parentMenuEl) {
      dom.setAttrs(parentMenuEl, {
        "menu-parent-proxy-id": this.autoTetheredMenuId
      });
      render(this.__renderMenu(), parentMenuEl);
    }
    return this.tetheredMenu;
  }
  __applyMenuSettings(menu) {
    if (!menu) {
      return;
    }
    const menuItems = f1.safe([], this.input);
    const hasChildren = f1.some((item) => Object.keys(item).includes(`${this.childrenAttr}`), menuItems);
    const { placement = "right", subtitle = "", capHeight = false } = f1.safe({}, this.menuSettings);
    dom.flipAttr(menu, "cap-height", capHeight && !hasChildren);
    dom.setAttrs(menu, { placement, subtitle });
  }
  __renderMenu() {
    return html`<hi-menu
      gid=${this.tetheredMenuId}
      id=${this.tetheredMenuId}
      .for=${this.forControl}
      .ns=${this.ns}
    ></hi-menu>`;
  }
  __renderMenuItems() {
    const menuItems = f1.safe([], this.input);
    const hasChildren = f1.some((item) => Object.keys(item).includes(`${this.childrenAttr}`), menuItems);
    if (hasChildren) {
      this.__findItemToApplyCapHeight(menuItems);
    }
    return html`${menuItems.map((_) => this.__renderPickOrList(_))}`;
  }
  __renderSubMenu(list) {
    const subMenuList = list[this.childrenAttr];
    return html`<hi-sub-menu .heading=${list.title} ?cap-height=${this.menuSettings?.capHeight && list.enableCapHeight}>
      ${f1.is.empty(subMenuList) ? nothing : subMenuList.map((_) => this.__renderPickOrList(_))}
    </hi-sub-menu>`;
  }
  __renderPickOrList(_) {
    const hasChildren = Object.keys(_).includes(`${this.childrenAttr}`);
    return html`${hasChildren ? this.__renderSubMenu(_) : this.__renderPick(_)}`;
  }
  __renderPick(pickList) {
    return html`<hi-pick .value=${pickList.value}>${pickList.title}</hi-pick>`;
  }
  __findItemToApplyCapHeight(items) {
    items.forEach((item) => {
      const child = item[this.childrenAttr];
      if (item.hasOwnProperty(`${this.childrenAttr}`) && child.length > 0) {
        const nestedChild = f1.filter((_) => Object.keys(_).includes(`${this.childrenAttr}`), child);
        if (nestedChild.length > 0) {
          this.__findItemToApplyCapHeight(nestedChild);
        } else {
          item.enableCapHeight = true;
        }
      }
    });
  }
  __menuControl__onOpenedChanged(e) {
    const isMenuClosed = !e.detail.value;
    if (isMenuClosed) {
      this.deactivate();
    }
  }
  __menuControl__repeatEvent(e) {
    this._dispatchEvent(e.type, e.detail);
  }
}
export default HIP.factory.element(MenuProxyElement);
