
import { stylesTheme } from '../../../eip/element/hi-component/hi-component-theme.js';
import { LitElement, css, html } from '../../../hip/library/lit-element/lit-element.js';
import { ref, createRef } from '../../../hip/library/lit-html/directives/ref.js';
import { AsLitHIC, HIP, BehaviorElement } from '../../../eip/element/hi-component/hi-component.js';
import ess from '../../../pip/library/pi-ess/pi-ess.js';
export const styles = css`
  :host {
    height: 100%;
    width: 100%;
    display: flex;
    align-items: center;
    padding: 0 var(--hi-px-8);
    background-color: var(--hi-clr--white);
  }

  .info {
    list-style-type: none;
    padding: 0;
    margin: 0;
  }

  #networkActivityIcon {
    border-radius: 1000px;
  }

  #networkActivityIcon.activity-on {
    animation: pulse-animation 2s infinite;
  }

  #networkActivityIcon.activity-once {
    animation: pulse-animation 2s 1;
  }

  @keyframes pulse-animation {
    0% {
      box-shadow: 0 0 0 0px rgba(0, 114, 206, 0.4);
    }
    70% {
      box-shadow: 0 0 0 0px rgba(0, 114, 206, 0.3);
    }
    100% {
      box-shadow: 0 0 0 1rem rgba(0, 114, 206, 0);
    }
  }
`;
export class NetworkActivityNavigationElement extends BehaviorElement(AsLitHIC(LitElement)) {
  constructor() {
    super();
    this.viewActivity = {
      transitioning: 0,
      transitioned: 0,
      errored: 0
    };
    this.networkActivity = {
      successfulConnections: 0,
      activeConnections: 0,
      failedConnections: 0
    };
    this.activityIconRef = createRef();
    this.networkLoading = false;
    this.viewLoading = false;
    this.updateDelay = 2e3;
    this.__onNetworkMessageBound = this.__onNetworkMessage.bind(this);
    this.__onViewMessageBound = this.__onViewMessage.bind(this);
  }
  static get is() {
    return "hi-navigation-network-activity";
  }
  static get styles() {
    return [stylesTheme, styles];
  }
  static get properties() {
    return {
      networkLoading: {
        type: Boolean,
        attribute: "network-loading"
      },
      viewLoading: {
        type: Boolean,
        attribute: "view-loading"
      },
      updateDelay: {
        type: Number,
        attribute: "update-delay"
      }
    };
  }
  connectedCallback() {
    super.connectedCallback();
    this.__networkActivityListener = ess.on(
      { site: "fi-hosting", source: "loading", feature: "network-activity" },
      this.__onNetworkMessageBound
    );
    this.__viewActivityListener = ess.on(
      { site: "fi-micro", source: "view_state", feature: "view-activity" },
      this.__onViewMessageBound
    );
  }
  disconnectedCallback() {
    super.disconnectedCallback();
    ess.off(this.__networkActivityListener);
    ess.off(this.__viewActivityListener);
  }
  render() {
    const ns = this.ns;
    const viewActivity = this.viewActivity;
    const networkActivity = this.networkActivity;
    return html`
      <hi-icon
        id="networkActivityIcon"
        gid="networkActivityIcon"
        ${ref(this.activityIconRef)}
        kind="wifi_tethering"
        type="material"
        .theme=${this.theme}
        .ns=${ns}
      ></hi-icon>
      <hi-tooltip for="networkActivityIcon" .ns=${ns}>
        <strong>Packets</strong>
        <ul class="info">
          <li>${networkActivity.activeConnections} currently loading</li>
          <li>${networkActivity.successfulConnections} successfully loaded</li>
          <li>${networkActivity.failedConnections} failed loading</li>
        </ul>
        <br />
        <strong>Views</strong>
        <ul class="info">
          <li>${viewActivity.transitioning} currently loading</li>
          <li>${viewActivity.transitioned} successfully loaded</li>
          <li>${viewActivity.errored} failed loading</li>
        </ul>
      </hi-tooltip>
    `;
  }
  __onNetworkMessage(event) {
    const { action, data } = event;
    this.__flipActivityIndicator(action);
    this.networkActivity = {
      successfulConnections: this.networkActivity.successfulConnections + data.successfulConnections,
      failedConnections: this.networkActivity.failedConnections + data.failedConnections,
      activeConnections: data.activeConnections
    };
  }
  __onViewMessage(event) {
    const { action, data } = event;
    this.__flipActivityIndicator(action);
    this.viewActivity = {
      transitioned: this.viewActivity.transitioned + data.transitioned,
      errored: this.viewActivity.errored + data.errored,
      transitioning: data.transitioning
    };
  }
  __flipActivityIndicator(action) {
    switch (action) {
      case "network-activity-started":
        this.networkLoading = true;
        this.__updateActivityIconStatus("activity-on");
        this.__resetActivityIndicators();
        break;
      case "view-activity-started":
        this.viewLoading = true;
        this.__updateActivityIconStatus("activity-on");
        this.__resetActivityIndicators();
        break;
      case "network-activity-stopped":
        this.networkLoading = false;
        this.__updateActivityIconStatus("activity-once");
        break;
      case "view-activity-stopped":
        this.viewLoading = false;
        this.__updateActivityIconStatus("activity-once");
        break;
    }
  }
  __resetActivityIndicators() {
    setTimeout(() => {
      this.__updateActivityIconStatus("");
      this.networkActivity.activeConnections = 0;
      this.viewActivity.transitioning = 0;
      this.requestUpdate("networkActivity");
      this.requestUpdate("viewActivity");
    }, this.updateDelay);
  }
  __updateActivityIconStatus(status) {
    const activityIconEl = this.activityIconRef.value;
    if (activityIconEl) {
      activityIconEl.setAttribute("class", status);
    }
  }
}
export default HIP.factory.element(NetworkActivityNavigationElement);
