
import { PolymerElement, html } from '../../../hip/library/polymer/polymer.js';
import { AsHIC, BehaviorElement, HIP } from '../../../eip/element/hi-component/hi-component.js';
import '../../../eip/element/hi-dialog/hi-dialog.js';
import '../../../eip/element/hi-button/hi-button.js';
import f1 from '../../../pip/library/pi-f1/pi-f1.js';
export class PromptElement extends BehaviorElement(AsHIC(PolymerElement)) {
  static get is() {
    return "hi-prompt";
  }
  static get template() {
    return html`
      <hi-dialog
        id="dialog"
        aesthetic="prompt"
        heading$="[[ title ]]"
        theme$="[[ theme ]]"
        scrollable
        disallow-close$="[[ disallowClose ]]"
        mode="sub-window"
      >
        <hi-icon id="statusMark" kind="success" size="large-xx" theme="green" slot="headerIcon"> </hi-icon>
        <p class="text-sz-m" inner-h-t-m-l="[[ details ]]"></p>
        <dom-repeat items="{{ controls }}">
          <template>
            <hi-button
              id$="[[ __actionButtonId(item) ]]"
              slot="controls"
              type$="[[ item.type ]]"
              aesthetic="[[ item.aesthetic ]]"
              on-trigger-ended="__onButtonTriggerEnded"
            >
              [[ item.name ]]
            </hi-button>
          </template>
        </dom-repeat>
      </hi-dialog>
    `;
  }
  static get properties() {
    return {
      heading: {
        type: String
      },
      text: {
        type: String
      },
      title: {
        type: String
      },
      details: {
        type: String
      },
      controls: {
        type: Array
      },
      closeControl: {
        type: Object
      },
      opened: {
        type: Boolean,
        reflectToAttribute: true,
        notify: true
      },
      disallowClose: {
        type: Boolean
      }
    };
  }
  static get instance() {
    return PromptElement.instances[PromptElement.instances.length - 1];
  }
  static get instances() {
    if (!PromptElement.__instances__) {
      PromptElement.__instances__ = [];
    }
    return PromptElement.__instances__;
  }
  connectedCallback() {
    super.connectedCallback();
    this.setAttribute("role", "group");
    PromptElement.instances.push(this);
    this.promptIndex = PromptElement.instances.length - 1;
  }
  disconnectedCallback() {
    super.disconnectedCallback();
    this.$.dialog.opened = false;
    PromptElement.instances.splice(this.promptIndex, 1);
  }
  _attachDom(dom) {
    this.appendChild(dom);
    return null;
  }
  __onButtonTriggerEnded(event) {
    const buttonData = event["model"].__data.item;
    this.remove();
    buttonData.promiseAction(buttonData);
  }
  __actionButtonId(item) {
    return `${item.type}Button`;
  }
  static create(configuration) {
    return HIP.fn.prompt(configuration);
  }
}
HIP.fn.prompt = (configuration) => {
  const controlCloseDefault = {
    value: -1,
    result: "reject"
  };
  const controlDefaults = {
    continue: {
      order: 3,
      aesthetic: "primary",
      type: "yes",
      name: "Continue",
      value: 1,
      result: "resolve"
    },
    yes: {
      order: 2,
      aesthetic: "primary",
      type: "yes",
      name: "Yes",
      value: 1,
      result: "resolve"
    },
    no: {
      order: 1,
      aesthetic: "secondary",
      type: "no",
      name: "No",
      value: 0,
      result: "reject"
    },
    cancel: {
      order: 0,
      aesthetic: "ghost",
      type: "cancel",
      name: "Cancel",
      value: -1,
      result: "reject"
    }
  };
  const successToIcon = {
    warning: "warning_filled",
    failure: "cross_filled",
    info: "light_bulb",
    success: "success"
  };
  const successToTheme = {
    warning: "orange",
    success: "green",
    failure: "red",
    info: "cyan"
  };
  configuration = configuration || {};
  configuration.controls = configuration.controls || [];
  return new Promise(function(resolve, reject) {
    const prompt = document.createElement("hi-prompt");
    const status = configuration.status;
    const results = {
      resolve,
      reject
    };
    document.body.appendChild(prompt);
    prompt.title = configuration.title || configuration.heading || "";
    prompt.details = configuration.details || configuration.text || "";
    prompt.disallowClose = f1.safe(true, configuration.disallow_close);
    prompt.closeControl = f1.merge("over", controlCloseDefault, f1.safe({}, configuration.close_control));
    prompt.$.statusMark.hidden = status == null;
    if (status != null) {
      prompt.$.statusMark.theme = successToTheme[status] || successToTheme["info"];
      prompt.$.statusMark.kind = successToIcon[status] || successToIcon["info"];
    }
    const closeButton = prompt.querySelector("hi-close");
    if (!prompt.disallowClose && closeButton) {
      closeButton.addEventListener(
        "trigger-ended",
        () => {
          const promiseFn = results[prompt.closeControl.result];
          promiseFn(prompt.closeControl.value);
        },
        {
          once: true
        }
      );
    }
    if (!prompt.disallowClose) {
      prompt.addEventListener(
        "keydown",
        (event) => {
          const promiseFn = results[prompt.closeControl.result];
          if (event.key === "Escape") {
            promiseFn(prompt.closeControl.value);
          }
        },
        {
          once: true
        }
      );
    }
    const compactFilter = (button) => !!button;
    const sortFn = (a, b) => a.order - b.order;
    prompt.controls = configuration.controls.map(function(_button) {
      let button;
      if (typeof _button === "string") {
        button = Object.assign({}, controlDefaults[_button]);
      } else if (typeof _button.type === "string") {
        button = Object.assign({}, controlDefaults[_button.type], _button);
      } else {
        button = _button;
      }
      if (!button.result) {
        prompt.warn(`(${button.name || button.value}) does not have a result key [resolve|reject]?`);
        return void 0;
      } else {
        button.promiseAction = results[button.result];
        return button;
      }
    }).filter(compactFilter).sort(sortFn);
    prompt.$.dialog.opened = true;
  });
};
HIP.fn.confirmation = (configuration) => {
  const confirmation = f1.merge(
    "over",
    {
      controls: ["cancel", "continue"],
      details: "This action cannot be reverted.<br/> Are you sure?",
      title: "Destructive Action",
      status: "warning"
    },
    configuration || {}
  );
  return HIP.fn.prompt(confirmation);
};
export default HIP.factory.element(PromptElement);
