import { Buffer } from 'buffer';
import { noChange, _$LH } from './lit-html.js';
import { PartType } from './directive.js';
import { isPrimitive, isTemplateResult, isSingleExpression } from './directive-helpers.js';

const { _TemplateInstance: TemplateInstance, _isIterable: isIterable, _resolveDirective: resolveDirective, _ChildPart: ChildPart, _ElementPart: ElementPart, } = _$LH;
const hydrate = (rootValue, container, options = {}) => {
    console.warn('Importing `hydrate()` from `../../../lit-html/experimental-hydrate.js` is deprecated.' +
        'Import from `@lit-labs/ssr-client` instead.');
    if (container['_$litPart$'] !== undefined) {
        throw new Error('container already contains a live render');
    }
    let rootPart = undefined;
    let rootPartMarker = undefined;
    let currentChildPart = undefined;
    const stack = [];
    const walker = document.createTreeWalker(container, NodeFilter.SHOW_COMMENT, null, false);
    let marker;
    while ((marker = walker.nextNode()) !== null) {
        const markerText = marker.data;
        if (markerText.startsWith('lit-part')) {
            if (stack.length === 0 && rootPart !== undefined) {
                throw new Error(`There must be only one root part per container. ` +
                    `Found a part marker (${marker}) when we already have a root ` +
                    `part marker (${rootPartMarker})`);
            }
            currentChildPart = openChildPart(rootValue, marker, stack, options);
            rootPart !== null && rootPart !== void 0 ? rootPart : (rootPart = currentChildPart);
            rootPartMarker !== null && rootPartMarker !== void 0 ? rootPartMarker : (rootPartMarker = marker);
        }
        else if (markerText.startsWith('lit-node')) {
            createAttributeParts(marker, stack, options);
        }
        else if (markerText.startsWith('/lit-part')) {
            if (stack.length === 1 && currentChildPart !== rootPart) {
                throw new Error('internal error');
            }
            currentChildPart = closeChildPart(marker, currentChildPart, stack);
        }
    }
    if (rootPart === undefined) {
        const elementMessage = container instanceof ShadowRoot
            ? `{container.host.localName}'s shadow root`
            : container instanceof DocumentFragment
                ? 'DocumentFragment'
                : container.localName;
        console.error(`There should be exactly one root part in a render container, ` +
            `but we didn't find any in ${elementMessage}.`);
    } 
    container['_$litPart$'] = rootPart;
};
const openChildPart = (rootValue, marker, stack, options) => {
    let value;
    let part;
    if (stack.length === 0) {
        part = new ChildPart(marker, null, undefined, options);
        value = rootValue;
    }
    else {
        const state = stack[stack.length - 1];
        if (state.type === 'template-instance') {
            part = new ChildPart(marker, null, state.instance, options);
            state.instance._$parts.push(part);
            value = state.result.values[state.instancePartIndex++];
            state.templatePartIndex++;
        }
        else if (state.type === 'iterable') {
            part = new ChildPart(marker, null, state.part, options);
            const result = state.iterator.next();
            if (result.done) {
                value = undefined;
                state.done = true;
                throw new Error('Unhandled shorter than expected iterable');
            }
            else {
                value = result.value;
            }
            state.part._$committedValue.push(part);
        }
        else {
            part = new ChildPart(marker, null, state.part, options);
        }
    }
    value = resolveDirective(part, value);
    if (value === noChange) {
        stack.push({ part, type: 'leaf' });
    }
    else if (isPrimitive(value)) {
        stack.push({ part, type: 'leaf' });
        part._$committedValue = value;
    }
    else if (isTemplateResult(value)) {
        const markerWithDigest = `lit-part ${digestForTemplateResult(value)}`;
        if (marker.data === markerWithDigest) {
            const template = ChildPart.prototype._$getTemplate(value);
            const instance = new TemplateInstance(template, part);
            stack.push({
                type: 'template-instance',
                instance,
                part,
                templatePartIndex: 0,
                instancePartIndex: 0,
                result: value,
            });
            part._$committedValue = instance;
        }
        else {
            throw new Error('Hydration value mismatch: Unexpected TemplateResult rendered to part');
        }
    }
    else if (isIterable(value)) {
        stack.push({
            part: part,
            type: 'iterable',
            value,
            iterator: value[Symbol.iterator](),
            done: false,
        });
        part._$committedValue = [];
    }
    else {
        stack.push({ part: part, type: 'leaf' });
        part._$committedValue = value == null ? '' : value;
    }
    return part;
};
const closeChildPart = (marker, part, stack) => {
    if (part === undefined) {
        throw new Error('unbalanced part marker');
    }
    part._$endNode = marker;
    const currentState = stack.pop();
    if (currentState.type === 'iterable') {
        if (!currentState.iterator.next().done) {
            throw new Error('unexpected longer than expected iterable');
        }
    }
    if (stack.length > 0) {
        const state = stack[stack.length - 1];
        return state.part;
    }
    else {
        return undefined;
    }
};
const createAttributeParts = (comment, stack, options) => {
    const match = /lit-node (\d+)/.exec(comment.data);
    const nodeIndex = parseInt(match[1]);
    const node = comment.nextElementSibling;
    if (node === null) {
        throw new Error('could not find node for attribute parts');
    }
    node.removeAttribute('defer-hydration');
    const state = stack[stack.length - 1];
    if (state.type === 'template-instance') {
        const instance = state.instance;
        while (true) {
            const templatePart = instance._$template.parts[state.templatePartIndex];
            if (templatePart === undefined ||
                (templatePart.type !== PartType.ATTRIBUTE &&
                    templatePart.type !== PartType.ELEMENT) ||
                templatePart.index !== nodeIndex) {
                break;
            }
            if (templatePart.type === PartType.ATTRIBUTE) {
                const instancePart = new templatePart.ctor(node, templatePart.name, templatePart.strings, state.instance, options);
                const value = isSingleExpression(instancePart)
                    ? state.result.values[state.instancePartIndex]
                    : state.result.values;
                const noCommit = !(instancePart.type === PartType.EVENT ||
                    instancePart.type === PartType.PROPERTY);
                instancePart._$setValue(value, instancePart, state.instancePartIndex, noCommit);
                state.instancePartIndex += templatePart.strings.length - 1;
                instance._$parts.push(instancePart);
            }
            else {
                const instancePart = new ElementPart(node, state.instance, options);
                resolveDirective(instancePart, state.result.values[state.instancePartIndex++]);
                instance._$parts.push(instancePart);
            }
            state.templatePartIndex++;
        }
    }
    else {
        throw new Error('internal error');
    }
};
const digestSize = 2;
const digestForTemplateResult = (templateResult) => {
    const hashes = new Uint32Array(digestSize).fill(5381);
    for (const s of templateResult.strings) {
        for (let i = 0; i < s.length; i++) {
            hashes[i % digestSize] = (hashes[i % digestSize] * 33) ^ s.charCodeAt(i);
        }
    }
    const str = String.fromCharCode(...new Uint8Array(hashes.buffer));
    return Buffer.from(str, 'binary').toString('base64') ;
};

export { digestForTemplateResult, hydrate };

//# sourceMappingURL=experimental-hydrate.js.map
