
import { html, LitElement, css, nothing } from '../../../hip/library/lit-element/lit-element.js';
import { AsLitHIC, BehaviorElement, HIP } from '../../../eip/element/hi-hip/hi-hif.js';
import { stylesTheme } from '../../../eip/element/hi-component/hi-component-theme.js';
import { stylesTile } from '../../../eip/element/hi-behavior-style/hi-style-tile.js';
import { Ignition } from '../../../pip/library/pi-interop/pi-interop.js';
import $ip from '../../../pip/library/pi-ip/pi-ip.js';
import f1 from '../../../pip/library/pi-f1/pi-f1.js';
import '../../../eip/element/hi-menu/hi-menu.js';
import '../../../eip/element/hi-button/hi-button.js';
import '../../../eip/element/hi-menu/hi-menu-proxy.js';
import '../../../yip/element/yi-view-header/yi-view-header.js';
import './vi-report-narrative.js';
export const styles = css`
  :host {
    display: block;
    padding: 0;
  }

  #header {
    display: flex;
    margin: 0;
  }

  #narrative-body {
    overflow: auto;
    height: calc(100% - var(--hi-el--height-lrg));
    width: 100%;
  }

  #narrative-info {
    margin-left: var(--hi-px-4);
  }

  @media screen and (max-width: 62rem) {
    :host {
      grid-column: 1 / -1 !important;
    }
  }
`;
export class NarrativeElement extends BehaviorElement(AsLitHIC(LitElement)) {
  constructor() {
    super(...arguments);
    this.ns = `${this.id}-ns`;
    this.willRender = false;
    this.pivotTableVersion = "23.6.1";
  }
  static get is() {
    return "vi-narrative";
  }
  static get styles() {
    return [stylesTheme, stylesTile, styles];
  }
  static get properties() {
    return {
      willRender: {
        type: Boolean
      },
      narrative: {
        type: Object
      },
      input: {
        type: Object
      },
      name: {
        type: String
      }
    };
  }
  disconnectedCallback() {
    super.disconnectedCallback();
    this.unloadPivotTable();
  }
  render() {
    if (!this.willRender) {
      return nothing;
    }
    const narrative = this.narrative;
    return html`
      ${this._renderHeader(narrative)}
      <section id="narrative-body">${narrative ? this._renderNarrativeBody(narrative, this.input) : nothing}</section>
    `;
  }
  _renderHeader(narrative) {
    const { title, description } = narrative;
    const ns = this.ns;
    const actions = narrative.actions;
    const hasActions = !f1.is.empty(actions);
    return html`
      <header id="header">
        <hi-title level="5" title=${title} .ns=${ns}></hi-title>
        ${description ? html`
              <hi-icon id="narrative-info" gid="narrative-info" .ns=${ns} glyph="info"></hi-icon>
              <hi-tooltip for="narrative-info" .ns=${ns}>${description}</hi-tooltip>
            ` : nothing}
        <span style="margin-left: auto"> ${hasActions ? this.__renderActionButton(narrative) : nothing}</span>
      </header>
    `;
  }
  __renderActionButton(narrative) {
    const actions = narrative.actions;
    const source = narrative.source;
    return html`<hi-button id="${source}_actions__btn" gid="${source}_actions__btn" aesthetic="icon" bordered>
        <hi-icon glyph="menu" type="material"></hi-icon>
      </hi-button>
      <hi-menu-proxy .input=${actions} for-control="${source}_actions__btn" for-menu="${source}_actions__menu">
      </hi-menu-proxy>
      <hi-menu
        gid="${source}_actions__menu"
        placement="left"
        @selected-ended=${(e) => this.__narrativeMenu__selectedEnded(e, narrative)}
      >
      </hi-menu>`;
  }
  _renderNarrativeBody(narrative, input) {
    const renderNarrative = $ip.di.inject("narrative", narrative.render_type);
    if (!renderNarrative) {
      return nothing;
    }
    const renderFn = renderNarrative.render(narrative, this);
    if (!renderFn) {
      return nothing;
    }
    return renderFn ? renderFn(input) : nothing;
  }
  firstUpdated(changedProperties) {
    super.firstUpdated(changedProperties);
    if (this.narrative && this.narrative.render_type === "pivot") {
      this.loadPivotTable();
    }
  }
  async loadPivotTable() {
    const { type, settings } = this.createFields(this.narrative, "pivot");
    const [pivotTable, cssLink] = await $ip.requires([
      `@crate/hip/pivot-table/${this.pivotTableVersion}/index`,
      `@crate/hip/pivot-table/${this.pivotTableVersion}/index.css`
    ]);
    this.pivotTable = pivotTable;
    await Ignition.create().startup(this.pivotTable, { version: this.pivotTableVersion });
    const pivotTableEl = this.shadowRoot?.querySelector("eab-hip-pivot-table");
    if (pivotTableEl) {
      pivotTableEl.settings = f1.merge(
        "over",
        {
          input: this.input,
          type,
          attached_to: pivotTableEl,
          event_to: pivotTableEl,
          report: {},
          options: {
            in_dialog: false
          }
        },
        settings
      );
      const style = document.createElement("style");
      style.textContent = `@import url('${cssLink.href}');`;
      this.shadowRoot?.appendChild(style);
    }
  }
  unloadPivotTable() {
    if (this.pivotTable) {
      Ignition.create().shutdown(this.pivotTable);
      $ip.unrequire(`hip/pivot-table/${this.pivotTableVersion}/index.css`);
    }
  }
  createFields(narrative, narrativeType) {
    const fields = f1.read("/fields", {}, narrative);
    const type = f1.read("/type", { fields }, narrative);
    const settings = f1.read("/settings", {}, narrative);
    const renderHints = f1.read("/render_hints", {}, narrative);
    f1.write("/render_hints", renderHints, type);
    return {
      id: `narrative_${narrativeType}_${f1.id()}`,
      attributes: f1.read("/attributes", {}, renderHints),
      style: f1.read("/style", {}, renderHints),
      type,
      settings
    };
  }
  __narrativeMenu__selectedEnded(evt, narrative) {
    this.narrative__onEventFired(evt, narrative, `${this.name}_action-ended`);
  }
  async narrative__onEventFired(evt, narrative, taskName = `${this.name}_${evt.type}`) {
    const id = narrative.id;
    const event = f1.is.object(evt.detail.value) ? evt.detail.value : evt.detail;
    this._dispatchNotification("narrative-updated", {
      taskName,
      narrative,
      event,
      id
    });
  }
}
export default HIP.factory.element(NarrativeElement);
