
import lucid from '../../../aip/library/ai-lucid/ai-lucid.js';
import f1 from '../../../pip/library/pi-f1/pi-f1.js';
const _ReportModel = class {
  constructor(ns) {
    this.ns = ns;
    this.runtime = lucid.createRuntime(this.ns);
    this.narrativeIds = [];
  }
  resolveRuntime() {
    const workspace = this.runtime.workspace;
    const front_matter = workspace.queryInstance("front_matter")?.value();
    const reportSettings = workspace.queryInstance("report")?.settings();
    const reportActions = f1.read("/actions", [], reportSettings);
    const narratives = f1.compact(
      workspace.queryInstances(
        f1.difference(["front_matter", "report"], this.narrativeIds),
        _ReportModel.constants.narrativeTaskTag
      )
    );
    return {
      report: {
        actions: reportActions
      },
      front_matter,
      data: f1.map((_) => _.value(), narratives),
      narratives: f1.map((_) => _.settings(), narratives)
    };
  }
  async generateRuntime(hrf) {
    this.runtime = lucid.createRuntime(this.ns);
    const sanitizedHrf = this.sanitize(f1.safe({}, hrf));
    const hrfHIL = this._hrfToHIL(sanitizedHrf);
    this.narrativeIds = f1.compact(f1.map(f1.read("/id", null), hrfHIL.startup));
    await this.runtime.computeWith(hrfHIL);
    return this;
  }
  _hrfToHIL(hrf) {
    const narratives = f1.read("/metadata/report/narratives", {}, hrf);
    const events = f1.read("/metadata/report/events", {}, hrf);
    const data = this._pullHRFDataSources(hrf);
    const tasks = f1.read("/metadata/report/tasks", {}, hrf);
    const startup = f1.to.array(f1.map((_, id) => id, tasks));
    if (!f1.is.in("front_matter", tasks)) {
      tasks["front_matter"] = {
        $is: "task",
        model: f1.read("/metadata/report/front_matter", {}, hrf)
      };
    }
    startup.push("front_matter");
    f1.each(
      (narrative, id) => {
        startup.push(id);
        tasks[id] = {
          $is: "task",
          tags: [_ReportModel.constants.narrativeTaskTag],
          settings: f1.clone(narrative),
          model: f1.clone(data[narrative.source])
        };
      },
      f1.filter((_, key) => !f1.is.in(key, tasks), narratives)
    );
    if (!f1.is.in("report", tasks)) {
      tasks["report"] = {
        $is: "task",
        model: null,
        settings: {
          actions: f1.read("/metadata/report/actions", [], hrf)
        }
      };
    }
    startup.push("report");
    f1.each((_, key) => {
      const eventTypes = f1.keys(_);
      eventTypes.forEach((event_type) => {
        const task_key = `${key}_${event_type}`;
        const model = f1.read("/model", {}, f1.read(`/${key}`, "", events));
        if (!f1.is.in(task_key, tasks)) {
          tasks[task_key] = {
            $is: "task",
            model,
            success: _[event_type]
          };
        }
      });
    }, events);
    const hil = {
      startup: f1.map((id) => ({ $is: "reference", id }), f1.unique(startup)),
      tasks
    };
    return hil;
  }
  _readHRFModels(hrf) {
    return f1.keys(f1.read("/metadata/report/narratives", {}, hrf));
  }
  _readHRFNarratives(hrf) {
    return f1.keys(f1.read("/metadata/report/narratives", {}, hrf));
  }
  _readHRFActions(hrf) {
    return f1.keys(f1.read("/metadata/report/actions", {}, hrf));
  }
  _readHRFEvents(hrf) {
    return f1.keys(f1.read("/metadata/report/events", {}, hrf));
  }
  _readHRFDataSources(hrf) {
    return f1.keys(f1.read("/data", {}, hrf));
  }
  _pullHRFDataSources(hrf) {
    const data = f1.read("/data", {}, hrf);
    return f1.is.array(data) ? { [_ReportModel.constants.defaultDataSource]: data } : data;
  }
  sanitize(hrf) {
    const reportUnsafe = f1.read(`/metadata/report`, {}, hrf);
    const reportSafe = f1.merge(
      "over",
      {
        front_matter: {},
        narratives: {},
        actions: [],
        layout: {}
      },
      reportUnsafe
    );
    reportSafe.narratives = f1.map((_, id) => this._sanitizeNarrative(_, id), reportSafe.narratives);
    reportSafe.front_matter = this._sanitizeFrontMatter(reportSafe.front_matter);
    reportSafe.actions = this._sanitizeActions(reportSafe.actions);
    return f1.write("/metadata/report", reportSafe, hrf);
  }
  _sanitizeFrontMatter(frontMatter) {
    const frontMatterSafe = f1.merge(
      "over",
      {
        id: f1.id(),
        title: "",
        description: "",
        final_note: "",
        level: 1,
        timestamp: Date.now(),
        associated_links: []
      },
      frontMatter
    );
    frontMatterSafe.associated_links = this._sanitizeActions(frontMatterSafe.associated_links);
    return frontMatterSafe;
  }
  _sanitizeNarrative(narrative, id) {
    const narrativeSafe = f1.merge(
      "over",
      {
        id,
        render_type: "fallback",
        title: f1.word(`${id}`),
        description: "",
        source: _ReportModel.constants.defaultDataSource,
        fields: {},
        render_hints: {},
        actions: [],
        layout: {
          row: 1,
          width: 12,
          height: 4
        }
      },
      narrative
    );
    narrativeSafe.render_hints = this._sanitizeRenderHint(narrativeSafe.render_hints);
    narrativeSafe.actions = this._sanitizeActions(narrativeSafe.actions);
    narrativeSafe.fields = f1.map((_) => this._sanitizeFields(_), narrativeSafe.fields);
    return narrativeSafe;
  }
  _sanitizeFields(field) {
    return field;
  }
  _sanitizeRenderHint(renderHint) {
    return renderHint;
  }
  _sanitizeActions(actions) {
    return f1.merge("over", [], actions);
  }
};
export let ReportModel = _ReportModel;
ReportModel.constants = {
  defaultDataSource: "report",
  narrativeTaskTag: "narrative"
};
