
import { html, LitElement, css, nothing } from '../../../hip/library/lit-element/lit-element.js';
import { styleMap } from '../../../hip/library/lit-html/directives/style-map.js';
import { AsLitHIC, BehaviorElement, HIP } from '../../../eip/element/hi-hip/hi-hif.js';
import { stylesTheme } from '../../../eip/element/hi-component/hi-component-theme.js';
import { ifDefined as ifDef } from '../../../hip/library/lit-html/directives/if-defined.js';
import f1 from '../../../pip/library/pi-f1/pi-f1.js';
import '../../../eip/element/hi-menu/hi-menu-proxy.js';
import '../../../yip/element/yi-view-header/yi-view-header.js';
import '../../../yip/element/yi-view-footer/yi-view-footer.js';
import '../../../eip/element/hi-title/hi-title.js';
import '../../../eip/element/hi-link/hi-link.js';
import { ReportModel } from './vi-report-model.js';
import './vi-narrative.js';
export const styles = css`
  :host {
    display: block;
    --grid-column-gap: 1em;
    --grid-row-gap: 1em;
    --grid-body-spacing: 2em;
  }

  .report-grid {
    display: grid;
    grid-template-columns: repeat(12, minmax(0, 1fr));
    grid-column-gap: var(--grid-column-gap);
    grid-row-gap: var(--grid-row-gap);
  }

  .report-header {
    margin-bottom: var(--grid-body-spacing);
  }

  .report-footer {
    margin-top: var(--grid-body-spacing);
  }
`;
export class ReportElement extends BehaviorElement(AsLitHIC(LitElement)) {
  constructor() {
    super(...arguments);
    this.model = new ReportModel(this.ns);
    this.__inputChanged = false;
  }
  static get is() {
    return "vi-report";
  }
  static get styles() {
    return [stylesTheme, styles];
  }
  static get properties() {
    return {
      input: {
        type: Object
      }
    };
  }
  connectedCallback() {
    super.connectedCallback();
    this.setAttribute("role", "group");
    this.setAttribute("gid", `force-paint ${this.getAttribute("gid")}`);
    this.__includeEventHandlers();
  }
  __includeEventHandlers() {
    this.addEventListener("narrative-updated-notified", this.__narrative__narrativeUpdatedNotified);
  }
  async update(changedProperties) {
    super.update(changedProperties);
    if (changedProperties.has("input")) {
      this.__inputChanged = true;
    }
  }
  async performUpdate() {
    super.performUpdate();
    if (this.__inputChanged) {
      await this.model.generateRuntime(this.input);
      await this.invokeReport(`report_initialize-ended`, {});
      this.requestUpdate();
      this.__inputChanged = false;
    }
  }
  render() {
    const internalModel = this.model.resolveRuntime();
    if (f1.is.empty(internalModel)) {
      return nothing;
    }
    const { front_matter, report, narratives, data } = internalModel;
    if (f1.is.empty(narratives) || f1.is.empty(data) || f1.is.empty(front_matter)) {
      return nothing;
    }
    const { title, level, description, final_note } = front_matter;
    const { actions } = report;
    const hasDescription = f1.is.exists(description);
    const hasReportActions = actions && actions.length > 0;
    return html`
      <yi-view-header class="report-header">
        <hi-title slot="title" level=${level} title=${title}></hi-title>
        ${hasDescription ? html` <p slot="description">${description}</p>` : nothing}
        ${hasReportActions ? html`
              <section slot="navigation">
                <hi-button id="report_header__actions" gid="report_header__actions" bordered aesthetic="icon">
                  <hi-icon glyph="menu" type="material"></hi-icon>
                </hi-button>
                <hi-menu-proxy
                  for-control="report_header__actions"
                  for-menu="report_actions__menu"
                  .input=${actions}
                  @selected-ended=${this.__reportAction__onSelectedEnded}
                >
                </hi-menu-proxy>
                <hi-menu gid="report_actions__menu" placement="left"></hi-menu>
              </section>
            ` : nothing}
      </yi-view-header>
      <div class="report-grid">${this._renderNarratives(narratives, data)}</div>
      <yi-view-footer class="report-footer">
        <p slot="description">${final_note}</p>
        <div slot="navigation">
          ${front_matter.associated_links.map(
      (action) => html`
              <hi-link href=${ifDef(action.value)} aesthetic="navigation">${action.title}</hi-link>
            `
    )}
        </div>
      </yi-view-footer>
    `;
  }
  updated() {
    if (!this.shadowRoot) {
      return;
    }
    const narrativeEls = this.shadowRoot.querySelectorAll("vi-narrative");
    if (!narrativeEls) {
      return;
    }
    narrativeEls.forEach((_) => this.__lazyLoadNarratives(_));
  }
  _renderNarratives(narratives, data) {
    const __columnPositions = {};
    const narrativeList = f1.to.array(narratives);
    const narrativeKeys = f1.keys(narratives);
    return f1.map((narrative, id) => {
      const gridStyles = this.__computeNarrativeLocation(narrative, __columnPositions);
      const narrativeKey = narrativeKeys[id];
      const narrativeData = f1.read(`/${narrative.id}`, void 0, data);
      return html`
        <vi-narrative
          id=${id}
          name=${narrativeKey}
          .input=${narrativeData}
          .narrative=${narrative}
          style=${styleMap(gridStyles)}
        ></vi-narrative>
      `;
    }, narrativeList);
  }
  __computeNarrativeLocation(narrative, __columnPositions) {
    const { row, width, height } = narrative.layout;
    const narrativeHeight = height * 100;
    const startColumn = __columnPositions[row] || 1;
    const endColumn = startColumn + width;
    __columnPositions[row] = endColumn;
    return {
      gridColumn: `${startColumn} / span ${width}`,
      height: `${narrativeHeight}px`
    };
  }
  __lazyLoadNarratives(target) {
    const intersectionObserver = new IntersectionObserver((entries, observer) => {
      entries.forEach((entry) => {
        if (entry.isIntersecting) {
          entry.target.willRender = true;
          observer.disconnect();
        }
      });
    });
    intersectionObserver.observe(target);
  }
  async __narrative__narrativeUpdatedNotified(evt) {
    const { taskName, event } = evt.detail;
    return this.invokeReport(taskName, event);
  }
  __reportAction__onSelectedEnded(evt) {
    return this.invokeReport(`report_action-ended`, evt.detail);
  }
  async invokeReport(taskName, event) {
    const instance = this.model.runtime.workspace.queryInstance(taskName);
    if (!instance) {
      return;
    }
    await instance.invoke({
      event
    });
    this.requestUpdate();
  }
}
export default HIP.factory.element(ReportElement);
