import { LitElement, html } from 'https://cdn.eab.com/toolkit/21-4-12/hip/library/lit-element/lit-element.js';

const getCurrentLocationParams = () => {
  return window.location.hash ? `?${window.location.hash.split('?')[1]}` : window.location.search;
};
const getURLParams = () => {
  const searchParams = getCurrentLocationParams();
  const formDataFromUrlParams = new URLSearchParams(searchParams);
  let paramsData = {};
  const VALID_PARAMS = ['token', 'question'];
  VALID_PARAMS.forEach((key) => {
    if (formDataFromUrlParams.get(key) != null) {
      paramsData[key] = formDataFromUrlParams.get(key);
    }
  });
  return paramsData;
};

class PrefabAppointmentFeedback extends HIP.BehaviorPrefab(HIP.BehaviorElement(LitElement)) {
  static get is() {
    return 'ai-prefab-appointment-feedback';
  }

  constructor() {
    super();
    this.questionNumber = 0;
    this.appointmentFeedbackResponseAnswers = [];
    this.currentQuestion = {};
    this.locationHashWatcher = () => {
      if (window.location.hash.indexOf('question') > -1) {
        let questionNumberFromUrl = getURLParams().question;
        this.questionNumber = questionNumberFromUrl - 1;
      }
      this.requestUpdate();
    };
  }

  connectedCallback() {
    super.connectedCallback();
    this.feedbackRequestData = this.getInstance('feedback_template').output.api_response;
    this.getFeedbackTemplate();

    this.getInstance('feedback_template_data').addEventListener('output-changed', (event) => {
      this.feedbackRequestData = this.getInstance('feedback_template_data').output;
      this.currentQuestion = this.feedbackRequestData.questions[this.questionNumber];
      this.requestUpdate();
    });
    window.addEventListener('hashchange', this.locationHashWatcher);
    window.addEventListener(
      'beforeunload',
      (event) => {
        event.preventDefault();
        return (event.returnValue = '');
      },
      {
        capture: true
      }
    );
  }

  getFeedbackTemplate() {
    let urlParams = getURLParams();
    this.getInstance('get_feedback_template').invoke({ token: urlParams.token });
  }

  disconnectedCallback() {
    super.disconnectedCallback();
    window.removeEventListener('hashchange', this.locationHashWatcher);
    window.removeEventListener('beforeunload', () => {});
  }

  getInstance(instanceName) {
    return HIP.db.gid.queryOne(instanceName, this.ns);
  }

  render() {
    if (!this.feedbackRequestData) return ``;

    return html`
      <ai-screen gid="_screen">
        <ai-state gid="_state__main" states="main">
          <div l-kind="vert">
            ${this.renderInstructions()}
            <hi-accordion class="mbl">
              <hi-bar>
                Appointment Details: ${this.feedbackRequestData.attendance.care_unit} |
                ${this.formatAttendanceDateTime()}
              </hi-bar>
              <hi-note>
                Appointment Date and Time: ${this.formatAttendanceDateTime()}
                <br />
                Organizer: ${this.feedbackRequestData.attendance.organizer}
                <br />
                Location: ${this.feedbackRequestData.attendance.location}
                <br />
                Topic: ${this.feedbackRequestData.attendance.topic}
                <br />
                ${this.renderMeetingTypeInfo()}
              </hi-note>
            </hi-accordion>
            ${this.renderProgressBar()}
            <div class="fluid-text" id="responseContainer">
              <div class="mbm">
                <h3 id="questionTitle" class="display-inline" tabindex="0">${this.currentQuestion.question_text}</h3>
                ${this.renderRequiredStarOrClearResponse()}
              </div>
              ${this.renderQuestion()} ${this.updateSavedValues()}
            </div>
            <div id="controlsContainer" class="mlz mtm mbl pos-relative">
              <div id="questionRequiredMessage" class="clr-red_30-txt mbm" hidden>
                Please answer the question to continue.
              </div>
              ${this.renderPreviousButton()} ${this.renderNextButton()} ${this.renderSubmitButton()}
            </div>
          </div>
        </ai-state>
      </ai-screen>
    `;
  }

  renderMeetingTypeInfo() {
    if (this.feedbackRequestData.attendance.meeting_type) {
      return html` Meeting Type: ${this.feedbackRequestData.attendance.meeting_type} `;
    }
  }

  renderInstructions() {
    if (this.feedbackRequestData.template.instructions) {
      return html`
        <hi-alert theme="info">
          Please Review the Following Information
          <span slot="details"> ${this.feedbackRequestData.template.instructions} </span>
        </hi-alert>
      `;
    }
  }

  renderProgressBar() {
    return html`
      <div id="progressBarContainer">
        <div class="mas">Progress</div>
        <hi-progress
          class="mtm mbm"
          min="0"
          max="${this.feedbackRequestData.questions.length}"
          value="${this.questionNumber + 1}"
        ></hi-progress>
      </div>
    `;
  }

  formatAttendanceDateTime() {
    const begin_date_time = new Date(this.feedbackRequestData.attendance.begin_time).toLocaleTimeString([], {
      year: 'numeric',
      month: 'numeric',
      day: 'numeric',
      hour: '2-digit',
      minute: '2-digit'
    });
    const end_time = new Date(this.feedbackRequestData.attendance.end_time).toLocaleTimeString([], {
      hour: '2-digit',
      minute: '2-digit'
    });

    //remove the "D" or "S" from the timezone to be standard with other time zone displays
    var time_zone = new Date(this.feedbackRequestData.attendance.end_time)
      .toLocaleTimeString([], { timeZoneName: 'short' })
      .split(' ')[2];
    time_zone = `${time_zone.split('')[0]}${time_zone.split('')[2]}`;

    return `${begin_date_time} - ${end_time} ${time_zone}`;
  }

  renderRequiredStarOrClearResponse() {
    if (JSON.parse(this.currentQuestion.required)) {
      return html` <sup class="display-inline clr-red_30-txt text-sz-l" aria-label="Required Question">*</sup>`;
    } else {
      return html` <hi-button aesthetic="ghost" @trigger-ended="${this.clearResponse}">
        <sub class="display-inline text-sz-l mls">
          <hi-icon glyph="clear"></hi-icon>
        </sub>
        <div class="display-inline text-sz-m clr-cyan_50-txt">Clear response</div>
      </hi-button>`;
    }
  }

  renderQuestion() {
    let urlHash = window.location.hash;
    if (urlHash.indexOf('question') > -1) {
      window.location.hash = urlHash.replace(/question.{1}\d{1}/, 'question=' + (this.questionNumber + 1));
    } else {
      window.location.hash = window.location.hash + '&question=' + (this.questionNumber + 1);
    }
    this.currentQuestion = this.feedbackRequestData.questions[this.questionNumber];
    this.questionId = this.currentQuestion.id;
    switch (this.currentQuestion.question_type) {
      case 'frequency':
        return html` <hi-radio-group id="radioButtonAnswer" group="radioButtons">
          <hi-repeat id="radioOptions" input='["1 Never", "2 Rarely", "3 Sometimes", "4 Often", "5 Frequently"]'>
            <template>
              <hi-radio value="[[ item ]]" class="mbm"></hi-radio>
              <hi-label class="mbm">[[ item ]]</hi-label>
            </template>
          </hi-repeat>
        </hi-radio-group>`;
        break;
      case 'nps':
        return html` <hi-radio-group id="radioButtonAnswer" group="radioButtons">
          <hi-repeat
            id="radioOptions"
            input='["0 Very unlikely", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10 Very likely"]'
          >
            <template>
              <hi-radio value="[[ item ]]" class="mbm" group="radioButtons"></hi-radio>
              <hi-label class="mbm">[[ item ]]</hi-label>
            </template>
          </hi-repeat>
        </hi-radio-group>`;
        break;
      case 'free_text':
        this.appointmentFeedbackResponseAnswers[this.questionNumber] = {
          answer: '',
          ...this.appointmentFeedbackResponseAnswers[this.questionNumber]
        };
        return html` <hi-textarea
          id="textareaAnswers"
          class="mbm"
          maxlength="8000"
          aria-label="free text answer area"
          value="${this.appointmentFeedbackResponseAnswers[this.questionNumber].answer || ''}"
        ></hi-textarea>`;
        break;
      case 'yes_no':
        return html` <hi-radio-group id="radioButtonAnswer" group="radioButtons">
          <hi-repeat id="radioOptions" input='["1 No", "2 Yes"]'>
            <template>
              <hi-radio value="[[ item ]]" class="mbm" group="radioButtons"></hi-radio>
              <hi-label class="mbm">[[ item ]]</hi-label>
            </template>
          </hi-repeat>
        </hi-radio-group>`;
        break;
      case 'select_many':
        return html` <hi-checkbox-group id="checkboxAnswers" group="checkBoxAnswer">
          <hi-repeat id="checkboxOptions" input=${JSON.stringify(this.currentQuestion.choices)}>
            <template>
              <hi-checkbox value="[[ item ]]" class="mbm"></hi-checkbox>
              <hi-label class="mbm">[[ item ]]</hi-label>
            </template>
          </hi-repeat>
        </hi-checkbox-group>`;
        break;
      case 'select_one':
        return html` <hi-radio-group id="radioButtonAnswer" group="radioButtons">
          <hi-repeat id="radioOptions" input=${JSON.stringify(this.currentQuestion.choices)}>
            <template>
              <hi-radio value="[[ item ]]" class="mbm" group="radioButtons"></hi-radio>
              <hi-label class="mbm">[[ item ]]</hi-label>
            </template>
          </hi-repeat>
        </hi-radio-group>`;
        break;
      case 'agreement':
        return html` <hi-radio-group id="radioButtonAnswer" group="radioButtons">
          <hi-repeat
            id="radioOptions"
            input='["1 Strongly disagree", "2 Disagree", "3 Neither agree nor disagree", "4 Agree", "5 Strongly Agree"]'
          >
            <template>
              <hi-radio value="[[ item ]]" class="mbm" group="radioButtons"></hi-radio>
              <hi-label class="mbm">[[ item ]]</hi-label>
            </template>
          </hi-repeat>
        </hi-radio-group>`;
        break;
      case 'quality':
        return html` <hi-radio-group id="radioButtonAnswer" group="radioButtons">
          <hi-repeat
            id="radioOptions"
            input='["1 Very poor", "2 Poor", "3 Neither poor nor good", "4 Good", "5 Very good"]'
          >
            <template>
              <hi-radio value="[[ item ]]" class="mbm" group="radioButtons"></hi-radio>
              <hi-label class="mbm">[[ item ]]</hi-label>
            </template>
          </hi-repeat>
        </hi-radio-group>`;
        break;
      case 'difficulty':
        return html` <hi-radio-group id="radioButtonAnswer" group="radioButtons">
          <hi-repeat
            id="radioOptions"
            input='["1 Very difficult", "2 Difficult", "3 Neither difficult nor easy", "4 Easy", "5 Very easy"]'
          >
            <template>
              <hi-radio value="[[ item ]]" class="mbm" group="radioButtons"></hi-radio>
              <hi-label class="mbm">[[ item ]]</hi-label>
            </template>
          </hi-repeat>
        </hi-radio-group>`;
        break;
      case 'satisfaction':
        return html` <hi-radio-group id="radioButtonAnswer" group="radioButtons">
          <hi-repeat
            id="radioOptions"
            input='["1 Very dissatisfied", "2 Dissatisfied", "3 Neither dissatisfied nor satisfied", "4 Satisfied", "5 Very Satisfied"]'
          >
            <template>
              <hi-radio value="[[ item ]]" class="mbm" group="radioButtons"></hi-radio>
              <hi-label class="mbm">[[ item ]]</hi-label>
            </template>
          </hi-repeat>
        </hi-radio-group>`;
        break;
    }
  }

  renderNextButton() {
    if (this.questionNumber < this.feedbackRequestData.questions.length - 1) {
      return html` <hi-button
        id="nextQuestionButton"
        aesthetic="primary"
        class="pos-relative to-right"
        @trigger-ended="${this.nextButtonLogic}"
      >
        Next
        <hi-icon class="mlm" glyph="chevron_right"></hi-icon>
      </hi-button>`;
    }
  }

  renderPreviousButton() {
    if (this.questionNumber > 0) {
      return html` <hi-button
        id="previousQuestionButton"
        aesthetic="secondary"
        @trigger-ended="${this.decrementQuestionNumber}"
      >
        <hi-icon glyph="chevron_left" class="mrm"></hi-icon>
        Previous Question
      </hi-button>`;
    }
  }

  renderSubmitButton() {
    if (this.questionNumber === this.feedbackRequestData.questions.length - 1) {
      return html` <hi-button
        id="submitQuestionsButton"
        aesthetic="primary"
        class="pos-relative to-right"
        @trigger-ended="${this.submitButtonLogic}"
      >
        Submit
      </hi-button>`;
    }
  }

  nextButtonLogic() {
    let checkboxAnswers = document.querySelector('#responseContainer #checkboxAnswers');
    let textareaAnswer = document.querySelector('#responseContainer #textareaAnswers');
    let radioButtonAnswer = document.querySelector('#responseContainer #radioButtonAnswer');
    let questionRequiredMessage = document.querySelector('#questionRequiredMessage');

    let answerValue;
    if (this.currentQuestion.question_type === 'select_many') {
      answerValue = checkboxAnswers.values;
    } else if (this.currentQuestion.question_type === 'free_text') {
      if (!textareaAnswer.value) {
        answerValue = '';
      } else {
        answerValue = textareaAnswer.value;
      }
    } else {
      answerValue = radioButtonAnswer.value;
    }
    if (JSON.parse(this.currentQuestion.required) && (!answerValue || answerValue === '' || answerValue.length === 0)) {
      questionRequiredMessage.removeAttribute('hidden');
      return;
    }
    questionRequiredMessage.setAttribute('hidden', 'true');
    let questionId = this.feedbackRequestData.questions[this.questionNumber].id;
    this.appointmentFeedbackResponseAnswers[this.questionNumber] = { question_id: questionId, answer: answerValue };

    this.clearResponse();
    this.questionNumber++;
    this.currentQuestion = this.feedbackRequestData.questions[this.questionNumber];
    this.requestUpdate();
  }

  decrementQuestionNumber() {
    this.questionNumber--;
    this.currentQuestion = this.feedbackRequestData.questions[this.questionNumber];
    this.requestUpdate();
  }

  updateSavedValues() {
    setTimeout(() => {
      let checkboxAnswers = document.querySelector('#responseContainer #checkboxAnswers');
      if (checkboxAnswers && this.appointmentFeedbackResponseAnswers[this.questionNumber]) {
        checkboxAnswers.values = this.appointmentFeedbackResponseAnswers[this.questionNumber].answer;
      }
      let radioButtonAnswer = document.querySelector('#responseContainer #radioButtonAnswer');
      if (radioButtonAnswer && this.appointmentFeedbackResponseAnswers[this.questionNumber]) {
        radioButtonAnswer.value = this.appointmentFeedbackResponseAnswers[this.questionNumber].answer;
      }
      let instructionsElement = document.querySelector('hi-alert span');
      if (instructionsElement) {
        instructionsElement.innerHTML = instructionsElement.innerHTML.replaceAll('\n', '<br>');
      }
    }, 100);
  }

  clearResponse() {
    let checkboxAnswers = document.querySelector('#responseContainer #checkboxAnswers');
    let textareaAnswer = document.querySelector('#responseContainer #textareaAnswers');
    let radioButtonAnswer = document.querySelector('#responseContainer #radioButtonAnswer');
    if (this.currentQuestion.question_type === 'select_many') {
      checkboxAnswers.values = '';
    } else if (this.currentQuestion.question_type === 'free_text') {
      textareaAnswer.value = '';
    } else {
      radioButtonAnswer.value = '';
    }
  }

  submitButtonLogic() {
    let checkboxAnswers = document.querySelector('#responseContainer #checkboxAnswers');
    let textareaAnswer = document.querySelector('#responseContainer #textareaAnswers');
    let radioButtonAnswer = document.querySelector('#responseContainer #radioButtonAnswer');
    let questionRequiredMessage = document.querySelector('#questionRequiredMessage');

    let answerValue;
    if (this.currentQuestion.question_type === 'select_many') {
      answerValue = checkboxAnswers.values;
    } else if (this.currentQuestion.question_type === 'free_text') {
      if (!textareaAnswer.value) {
        answerValue = '';
      } else {
        answerValue = textareaAnswer.value;
      }
    } else {
      answerValue = radioButtonAnswer.value;
    }
    if (JSON.parse(this.currentQuestion.required) && (!answerValue || answerValue === '' || answerValue.length === 0)) {
      questionRequiredMessage.removeAttribute('hidden');
      return;
    }
    questionRequiredMessage.setAttribute('hidden', 'true');
    let questionId = this.feedbackRequestData.questions[this.questionNumber].id;
    this.appointmentFeedbackResponseAnswers[this.questionNumber] = { question_id: questionId, answer: answerValue };

    let urlParams = getURLParams();
    this.getInstance('submit_feedback_template').invoke({
      token: urlParams.token,
      answers: this.appointmentFeedbackResponseAnswers
    });
  }
}

HIP.factory.element(PrefabAppointmentFeedback);
