import { LitElement, html } from 'https://cdn.eab.com/toolkit/21-4-12/hip/library/lit-element/lit-element.js';
import { repeat } from 'https://cdn.eab.com/toolkit/21-4-12/hip/library/lit-html/directives/repeat.js';
import { unsafeHTML } from 'https://cdn.eab.com/toolkit/21-4-12/hip/library/lit-html/directives/unsafe-html.js';
// import { result_template } from './result-template.js';

// import {
//   LitElement,
//   html
// } from 'lit-element';
// const {
//   repeat
// } = HIP.extern.lit.directives;

// const repeat = HIP.extern.lit.directives.repeat;

/**
 * TODO:
 * 1. Remove dependency on CDN for lit
 * 3. Refactor templates into separate files
 * 6. Optimize LitElement for DOM performance
 * 7. Fix AX issues
 */

// PURE FUNCTIONS

/**
 * STC implememtatiom of Rails uses custom format of multi param serialization for queryparams
 * @example
 *
 * ```js
 *  //Standard browser serialization for query params
 *  ?a=1&b=2&b=3
 *  // Rails needs
 *  ?a=1&b[]=2&b[]=3
 * ```
 */
const transformRailsCompatibleParams = (params) => {
  for (const k in params) {
    if (Array.isArray(params[k])) {
      params[`${k}[]`] = params[k];
      delete params[k];
    }
  }
  return params;
};

const getURLParams = () => {
  const searchParams = window.location.hash ? `?${window.location.hash.split('?')[1]}` : window.location.search;
  const formDataFromUrlParams = new URLSearchParams(searchParams);
  const formKeys = [
    'care_unit_id',
    'service_id',
    'begin_date',
    'staff_ids',
    'meeting_type_id',
    'location_id',
    'course_id',
    'invitation_id',
    'token',
    'pal_guid',
    'staff_at_location',
    'location_page'
  ];
  let paramsData = {};
  formKeys.forEach((key) => {
    if (formDataFromUrlParams.has(key)) {
      const paramValues = formDataFromUrlParams.getAll(key);
      if (paramValues.length > 1) {
        paramsData[key] = paramValues.map((v) => parseInt(v, 10));
      } else {
        if (paramValues[0].match(/\D/)) {
          paramsData[key] = paramValues[0];
        } else {
          paramsData[key] = parseInt(paramValues[0], 10);
        }
      }
    }
  });
  return paramsData;
};

const formatAvailabilityTime = (time, duration) => {
  return `${moment(time).format('h:mm')} - ${moment(time).add(duration, 'minutes').format('h:mm A')}`;
};

const formatRangeTime = (begin_time, end_time) => {
  return `${moment(begin_time).format('h:mm A')} - ${moment(end_time).format('h:mm A')}`;
};

const formatDateToHumanReadableString = (date) => {
  return moment(date).format('ddd, MMM Do');
};

const timesTemplate = (staffTimes, duration, selectionCallback) => {
  const LIMIT = 15;
  const totalTimes = staffTimes.length;
  const limitedSet = staffTimes.slice(0, LIMIT);
  const rest = staffTimes.slice(LIMIT);
  const timesRepeat = (times) => {
    return html`${repeat(times, (time, index) => {
      return html`<hi-button
        size="small"
        @trigger-ended="${selectionCallback}"
        aesthetic="secondary"
        class="fixed-width-times-button bdr-round-full mrsx mbsx"
        value="${time}"
        >${formatAvailabilityTime(time, duration)}</hi-button
      >`;
    })}`;
  };

  const noTimesTemplate = html`<p class="text-center bdr-dashed bas clr-gray_50-bdr clr-gray_70-txt pam">
    No availability for this day
  </p>`;
  const fewTimesTemplate = html`${timesRepeat(staffTimes)}`;
  const manyTimeTemplate = html`
    <div>
      <hi-reveal kind="additional">
        ${timesRepeat(limitedSet)}
        <span slot="additional-content"> ${timesRepeat(rest)} </span>
      </hi-reveal>
    </div>
  `;
  if (totalTimes > LIMIT) {
    return manyTimeTemplate;
  } else if (totalTimes > 0) {
    return fewTimesTemplate;
  } else {
    return noTimesTemplate;
  }
};

class PrefabAppointmentResultsElement extends HIP.BehaviorPrefab(HIP.BehaviorElement(LitElement)) {
  static get is() {
    return 'ai-prefab-appointment-results';
  }

  constructor() {
    super();
    this.renderRoot = this;
    this.mobileFiltersToggle = false;
    this.filteredResults = [];
    this.allMappedResults = [];
    this.filteredFormDefinition = {};
    this.urlParams = {};
    this.apiParams = {};
    this.selectedDate = this.__getCurrentDateFormatted;
    this.__availableDates = {};
    this.datesRequested = {};
    this.isFetchingResults = true;
    this.viewIndividualAvailabilitiesForLocation = 0;
    this.is_requestable = false;
    this.is_visitable = false;
    this.is_invitation = false;
    this.is_unauthenticated = false;
    this.invitationId;
    this.token;
    this.palGuid;

    this.locationHashWatcher = (hasChangeEvent) => {
      let courseIdBeforeChange = this.urlParams['course_id'];
      let locationIdBeforeChange = this.urlParams['location_id'];
      let meetingTypeIdBeforeChange = this.urlParams['meeting_type_id'];
      this.urlParams = getURLParams();
      let needsRefresh;

      // Check if date is outside of date range
      if (!this.is_requestable && !this.is_visitable) {
        needsRefresh = !this.datesRequested[this.urlParams.begin_date];
      } else {
        needsRefresh = locationIdBeforeChange != this.urlParams['location_id'];
      }

      // If course_id param changed, we need to refresh data.
      if (courseIdBeforeChange != this.urlParams['course_id']) {
        needsRefresh = true;
      }

      // If meeting_type_id param changed, we need to refresh data.
      if (meetingTypeIdBeforeChange != this.urlParams['meeting_type_id']) {
        needsRefresh = true;
      }

      this.viewIndividualAvailabilitiesForLocation = 0;

      // If every param we sent to api is not currently in url params, we need to refresh data.
      Object.keys(this.apiParams).forEach((paramName) => {
        if (!this.urlParams[paramName]) {
          needsRefresh = true;
        }
      });
      if (needsRefresh) {
        this.getResults(this.urlParams);
      } else {
        this.__updateFiltersAndResultsView();
      }
      if (this.urlParams.begin_date) {
        this.selectedDate = this.urlParams.begin_date;
      }
    };
  }

  connectedCallback() {
    super.connectedCallback();

    const contextOutput = this.getInstance('context').output;
    this.is_requestable = !!contextOutput.is_requestable;
    this.is_visitable = !!contextOutput.is_visitable;
    this.is_invitation = !!contextOutput.is_invitation;
    this.is_unauthenticated = !!contextOutput.is_unauthenticated;
    this.urlParams = getURLParams();
    if (this.is_invitation && this.urlParams.invitation_id) {
      this.invitationId = this.urlParams.invitation_id;
    }
    if (this.is_unauthenticated && this.urlParams.token) {
      this.token = this.urlParams.token;
    }
    if ('pal_guid' in this.urlParams) {
      this.palGuid = this.urlParams.pal_guid;
    }
    if (this.urlParams.begin_date) {
      this.selectedDate = this.urlParams.begin_date;
    } else if (!this.is_requestable && !this.is_invitation) {
      this.urlParams.begin_date = this.selectedDate;
    }

    this.getInstance('results').addEventListener('output-changed', (event) => {
      setTimeout(() => {
        this.__updateFiltersAndResultsView();

        // If we get data back, update this.datesRequested to include 30 day range
        let dateToAdd = moment(this.urlParams.begin_date);
        for (let i = 0; i < 30; i++) {
          this.datesRequested[dateToAdd.format('YYYY-MM-DD')] = true;
          dateToAdd.add(1, 'day');
        }
        const that_schedulable_data = this.getInstance('base_schedulable_data');
        if (that_schedulable_data !== undefined && that_schedulable_data.output === undefined) {
          that_schedulable_data.addEventListener('output-changed', (event) => {
            this.requestUpdate(); // Lit-Element native method.
          });

          this.getSchedulables();
        }

        const that_requestable_data = this.getInstance('base_requestable_data');
        if (that_requestable_data !== undefined && that_requestable_data.output === undefined) {
          that_requestable_data.addEventListener('output-changed', (event) => {
            this.requestUpdate(); // Lit-Element native method.
          });

          this.getRequestables();
        }

        const that_visitable_data = this.getInstance('base_visitable_data');
        if (that_visitable_data !== undefined && that_visitable_data.output === undefined) {
          that_visitable_data.addEventListener('output-changed', (event) => {
            this.requestUpdate(); // Lit-Element native method.
          });

          this.getVisitables();
        }
      }, 0);

      const custom_content_bucket = this.getInstance('custom_content');
      if (custom_content_bucket !== undefined) {
        custom_content_bucket.addEventListener('output-changed', (event) => {
          this.__updateFiltersAndResultsView();
          this.requestUpdate();
        });
      }
      //Call Custom Text API if data not available
      const customTextLocalStore = this.getInstance('store_custom_content');
      const getCustomTextAPI = this.getInstance('get_custom_content');
      if (customTextLocalStore && customTextLocalStore.output == null) {
        const customContentParams = {};
        if (this.token) {
          customContentParams['token'] = this.token;
        }
        getCustomTextAPI.invoke(customContentParams);
      }
    });

    this.getInstance('get_appointments').addEventListener('output-changed', () => {
      this.isFetchingResults = false;
    });

    // Watch location changes and update.
    window.addEventListener('hashchange', this.locationHashWatcher);

    this.getResults(this.urlParams);
  }

  disconnectedCallback() {
    super.disconnectedCallback();
    window.removeEventListener('hashchange', this.locationHashWatcher);
  }

  static get properties() {
    return {
      /**
       * all links are direct and not altered to work as relative links (client-side)
       */
      isDirect: Boolean,
      /**
       * the navigation bucket
       */
      index: Object
    };
  }

  get defaultScript() {
    return {
      system: {
        state: 'main'
      },
      action: {
        on_filter_update: {}
      },
      bucket: {
        context: {},
        filters: {},
        results: {}
      },
      io: {
        get_appointments: {}
      }
    };
  }

  render() {
    return html`<ai-screen gid="_screen" ns="${this.ns}">
      ${this.is_invitation || this.is_unauthenticated ? this.__alertTemplate() : ``}
      <ai-state gid="_state__main" ns="${this.ns}" states="main" class="" l-kind="hort">
        <div
          class="appointment_filters ptm max-w-360 mrmx prmx brs bdr-solid clr-gray_20-bdr ${this.mobileFiltersToggle
            ? 'mobile-open'
            : ''}"
        >
          <!-- Filter Form -->
          <div
            class="appointment_filters__header"
            l-kind
            l-width-fixed
            l-justify="between"
            l-align="center"
            l-size="grow"
          >
            <h2 class="clr-gray_80-txt">All Filters</h2>
            ${this.is_unauthenticated || this.is_invitation
              ? ``
              : html`<hi-link href="#" class="text-right" @trigger-ended="${this.__clearAll}">Start Over</hi-link>`}
          </div>
          <hi-form
            ns="${this.ns}"
            id="_filters_form"
            gid="_filters_form"
            aesthetic="invisible"
            flush
            disable-http=""
            kind="filter"
            @output-changed="${this.__filterResultsOnFormUpdate}"
          >
            <ai-builder-form gid="_filters_form_builder" ns="${this.ns}" .input=${this.filteredFormDefinition}>
            </ai-builder-form>
          </hi-form>
        </div>
        <div class="appointment_results ptm ${this.mobileFiltersToggle ? 'mobile-open' : ''}">
          ${this.isFetchingResults ? `` : this.__selectionTagsTemplate()}
          <!-- Appointment Results -->
          <div class="pamx">
            ${this.__appointmentResults(this.filteredResults)}
            ${this.__escapeHatchTemplate(
              this.__seeSchedulableTimesTemplate(),
              this.__requestAlternateTimeTemplate(),
              this.__seeOtherDropinTimesTemplate(),
              this.__palEscapeTemplate()
            )}
          </div>
        </div>
      </ai-state>
    </ai-screen> `;
  }

  // UTILITY FNS
  get __getCurrentDateFormatted() {
    return moment().format('YYYY-MM-DD');
  }

  getInstance(instanceName) {
    return HIP.db.gid.queryOne(instanceName, this.ns);
  }

  getDefinition(bucketName) {
    const bucket = this.getInstance(bucketName);
    if (bucket) {
      return this.getInstance(bucketName).output;
    }
  }

  getResults(URLParams) {
    this.isFetchingResults = true;
    this.apiParams = { ...this.urlParams };
    this.datesRequested = {};
    if (this.is_invitation) {
      this.apiParams = {
        invitation_id: this.apiParams.invitation_id,
        begin_date: this.apiParams.begin_date,
        meeting_type_id: this.apiParams.meeting_type_id
      };
    }
    this.getInstance('get_appointments').invoke(transformRailsCompatibleParams({ ...this.apiParams }));

    this.requestUpdate(); // Lit-Element native method.
  }

  // OPERATING LOGIC

  __updateFiltersAndResultsView() {
    const fullResults = tt.deepClone(this.results);
    const filteredResults = this.__filterResults(this.urlParams, fullResults);
    const mappedResults = this.__mapResultsData(filteredResults);
    this.updateFilterDef(this.urlParams, mappedResults);
    this.filteredResults = mappedResults.resources;
    this.requestUpdate(); // Lit-Element native method.
  }

  __mapResultsData(results) {
    let dateOptions = [];
    let dedupedResources = this.dedupeResources(results.resources);

    results.resources = dedupedResources.map((resource) => {
      resource.location = results.locations.find((lc) => lc.id == resource.location_id) || {};
      resource.service = results.services.find((srv) => srv.id == resource.service_id) || {};
      resource.careUnit = results.care_units.find((cu) => cu.id == resource.care_unit_id) || {};

      if (resource.staff_times) {
        resource.staffs = results.staff.filter((s) => resource.staff_times.some((st) => s.id === st.staff_id));
      } else if (resource.dropin_times) {
        //Preserve the sorting order
        resource.dropin_times = results.staff
          .map((staff) => resource.dropin_times.find((dt) => dt.staff_id == staff.id))
          .filter((item) => !!item);

        resource.dropin_times.forEach((dt) => (dt.staff = results.staff.find((s) => s.id == dt.staff_id)));
      }

      if (resource.staff_times) {
        // Populates resource.times from resources.staff_times[].times[]
        this.__groupStaffTimesForLocation(resource);
        dateOptions = dateOptions.concat(resource.times);
        // Populates resource.timesByDate on each resource based on selected date
        if (this.urlParams.begin_date) {
          resource = this.__filterByDate(resource, this.urlParams.begin_date);
        }
      }

      return resource;
    });

    dateOptions = Array.from(new Set(dateOptions)).map((o) => o.split('T')[0]);
    if ('schedulable_dates' in results) {
      results.schedulable_dates = results.schedulable_dates.filter((o) => dateOptions.includes(o));
    }

    return results;
  }

  dedupeResources(resources) {
    const seenServiceLocations = {};

    return resources.filter((resource) => {
      let key = `${resource.service_id}-${resource.location_id}`;

      return seenServiceLocations[key] ? false : (seenServiceLocations[key] = true);
    });
  }

  updateFilterDef(formData = {}, results) {
    const def = this.getDefinition('filters_type');
    const care_units = results.care_units || [];
    const services = results.services || [];
    const staff = results.staff || [];
    const locations = results.locations || [];
    const meeting_types = results.meeting_types || [];
    const courses = results.courses || [];
    const availableTimes = results.schedulable_dates || [];
    const customContentText = this.getDefinition('custom_content');
    this.__availableDates = availableTimes;

    if (care_units.length === 1) formData.care_unit_id = care_units[0].id;
    if (services.length === 1) formData.service_id = services[0].id;

    // TODO: Commenting out these two lines seems to fix a lot of issues.
    // if (staff.length === 1) formData.staff_ids = staff[0].id;
    // if (locations.length === 1) formData.location_id = locations[0].id;

    /**
     * Below code is mapping dropdown values & setting picked options from URL.
     */
    if (def && def.properties) {
      // Show single choice as user should not update it
      def.properties.care_unit_id.options = care_units;
      if ('care_unit_id' in formData) {
        def.properties.care_unit_id.options = care_units.filter((c) => c.id == formData.care_unit_id);
        def.properties.care_unit_id.initial_value = [formData.care_unit_id];
      }
      // Show single choice as user should not update it
      const filteredServices = formData.care_unit_id
        ? services.filter((s) => s.care_unit_id == formData.care_unit_id)
        : [];
      def.properties.service_id.options = filteredServices;
      if ('service_id' in formData) {
        def.properties.service_id.initial_value = [formData.service_id];
        def.properties.service_id.options = filteredServices.filter((s) => s.id == formData.service_id);
      }

      if (def.properties.staff_ids) {
        staff.forEach((user) => {
          if (user.relationships.length > 0) {
            const last_relationship_type = user.relationships.slice(-1)[0];
            const all_other_relationship_types = user.relationships.slice(0, -1);
            if (all_other_relationship_types.length == 0) {
              user.title = `${user.title} (Your ${user.relationships[0]})`;
            } else {
              user.title = `${user.title} (Your ${all_other_relationship_types.join(
                ', '
              )} and ${last_relationship_type})`;
            }
          }
        });

        def.properties.staff_ids.options = staff.map((s) => {
          s.value = parseInt(s.value, 10);
          return s;
        });
        if ('staff_ids' in formData) {
          if (Array.isArray(formData.staff_ids)) {
            def.properties.staff_ids.initial_value = formData.staff_ids.map((v) => parseInt(v, 10));
          } else {
            def.properties.staff_ids.initial_value = [parseInt(formData.staff_ids)];
          }
        } else {
          def.properties.staff_ids.initial_value = [];
        }
      }

      def.properties.location_id.options = locations;
      'location_id' in formData
        ? (def.properties.location_id.initial_value = [formData.location_id])
        : (def.properties.location_id.initial_value = []);
      def.properties.staff_at_location.options = locations;

      if (def.properties.meeting_type_id) {
        def.properties.meeting_type_id.options = meeting_types;
        def.properties.meeting_type_id.initial_value =
          formData.hasOwnProperty('meeting_type_id') && formData.meeting_type_id ? [formData.meeting_type_id] : [];
      }

      def.properties.course_id.options = courses.map((course) => {
        course.title = `${course.name} (${course.number})`;
        return course;
      });
      'course_id' in formData
        ? (def.properties.course_id.initial_value = [formData.course_id])
        : (def.properties.course_id.initial_value = []);

      if (formData.begin_date && def.properties.begin_date && def.properties.begin_date.attributes) {
        if (results.campaign) {
          const beginDate = moment(results.campaign.start_at);
          const currentDate = moment();
          const selectedDate = moment(formData.begin_date);
          const currentMonth = currentDate.month();
          const selectedMonth = selectedDate.month();
          const minDate = beginDate.isBefore(currentDate) ? currentDate : beginDate;
          const minDateFormatted = minDate.format('YYYY-MM-DD');
          def.properties.begin_date.attributes.min = minDateFormatted;
          def.properties.begin_date.attributes.max = moment(results.campaign.end_at).format('YYYY-MM-DD');
          if (selectedDate.isBefore(minDate)) {
            def.properties.begin_date.initial_value = minDateFormatted;
          } else {
            def.properties.begin_date.initial_value = selectedDate.format('YYYY-MM-DD');
          }
          if (currentMonth === selectedMonth) {
            def.properties.begin_date.attributes['disabled-below-min'] = true;
          } else {
            delete def.properties.begin_date.attributes.min;
            delete def.properties.begin_date.attributes['disabled-below-min'];
          }
        } else {
          const selectedMonth = moment(formData.begin_date).month();
          const currentMonth = moment().month();
          def.properties.begin_date.attributes['disabled-below-min'] = currentMonth === selectedMonth;
          def.properties.begin_date.initial_value = moment(formData.begin_date).format('YYYY-MM-DD');
        }
        def.properties.begin_date.attributes['highlight-dates'] = JSON.stringify(
          availableTimes.map((el) => {
            if (el.includes('-')) {
              return el;
            } else {
              // Restore the dashes if the server did not give us dashes
              return `${el.substring(0, 4)}-${el.substring(4, 6)}-${el.substring(6, 8)}`;
            }
          })
        );
      }

      if ('invitation_id' in formData) {
        def.properties.invitation_id.inital_value = this.invitationId;
        def.properties.invitation_id.value = this.invitationId;
      }

      if ('token' in formData) {
        def.properties.token.inital_value = this.token;
      }

      if ('pal_guid' in formData && 'pal_guid' in def.properties) {
        def.properties.pal_guid.inital_value = formData.pal_guid;
        def.properties.pal_guid.value = formData.pal_guid;
      }

      this.__updateCustomLabels(def);

      if (def.properties.meeting_type_id) {
        if (def.properties.meeting_type_id.options.length === 0) {
          def.properties.meeting_type_id.hidden = true;
        }
      }
      this.filteredFormDefinition = { ...def };
    }

    if (getURLParams().location_page) formData.location_page = getURLParams().location_page;

    this.setURLParams(formData);
  }

  __updateCustomLabels(def) {
    // Udpate custom labels
    const customContentText = this.getDefinition('custom_content');
    def.properties.location_id.title =
      customContentText.student_scheduling.location_label || def.properties.location_id.title;
    def.properties.care_unit_id.title =
      customContentText.student_scheduling.care_unit_label || def.properties.care_unit_id.title;
    def.properties.service_id.title =
      customContentText.student_scheduling.service_label || def.properties.service_id.title;
    def.properties.course_id.title =
      customContentText.student_scheduling.course_label || def.properties.course_id.title;
    if (def.properties.staff_ids) {
      def.properties.staff_ids.title =
        customContentText.student_scheduling.staff_label || def.properties.staff_ids.title;
    }

    def = { ...def };
  }

  __filterResultsOnFormUpdate(e) {
    if (this.urlParams.staff_at_location !== undefined) {
      return;
    }
    const formData = e.detail.value;
    if (this.invitationId) {
      formData.invitation_id = this.invitationId;
    }
    if (this.token) {
      formData.token = this.token;
    }
    if (this.palGuid) {
      formData.pal_guid = this.palGuid;
    }

    if (this.urlParams.location_page) {
      formData.location_page = this.urlParams.location_page;
    }
    this.setURLParams(formData);
  }

  __filterResults(selectedFilters = this.urlParams, results) {
    /**
     * Filter data based on form values, Form values & results should have the same keys.
     */

    let staffOptions = [];
    let locationOptions = [];
    let meetingTypeOptions = [];
    let courseOptions = [];
    let timeOptions = [];

    // filter out resources that dn't match params
    results.resources = results.resources.filter((resource) => {
      const staffIds = Array.isArray(selectedFilters.staff_ids)
        ? selectedFilters.staff_ids.map((v) => parseInt(v, 10))
        : [parseInt(selectedFilters.staff_ids)];

      // Filter out resource if it doesn't have selected location_ids
      if (selectedFilters.location_id && selectedFilters.location_id !== resource.location_id) {
        return false;
      }

      if (resource.staff_times) {
        resource.staff_times = resource.staff_times.filter((staff_time) => {
          // Filter out staff_time if it doesn't have selected course_id
          if (selectedFilters.course_id && !staff_time.course_ids.includes(selectedFilters.course_id)) {
            return false;
          }

          // Filter out staff_time if it doesn't have selected staff_ids
          if (selectedFilters.staff_ids && !staffIds.includes(staff_time.staff_id)) {
            return false;
          }

          staffOptions.push(staff_time.staff_id);
          courseOptions = courseOptions.concat(staff_time.course_ids);
          meetingTypeOptions = meetingTypeOptions.concat(staff_time.meeting_type_ids);

          return true;
        });

        // Filter out resource if it has no staff_times
        if (resource.staff_times.length === 0) {
          return false;
        } else {
          timeOptions.concat(resource.staff_times);
        }
      }

      if (resource.dropin_times) {
        resource.dropin_times = resource.dropin_times.filter((staff_time) => {
          // Filter out staff_time if it doesn't have selected course_id
          if (selectedFilters.course_id && !staff_time.course_ids.includes(selectedFilters.course_id)) {
            return false;
          }

          staffOptions.push(staff_time.staff_id);

          // Filter out staff_time if it doesn't have selected staff_ids
          if (selectedFilters.staff_ids && !staffIds.includes(staff_time.staff_id)) {
            return false;
          }

          return true;
        });

        // Filter out resource if it has no dropin_times
        if (resource.dropin_times.length === 0) {
          return false;
        }
      }

      locationOptions.push(resource.location_id);

      return true;
    });

    if ('staff' in results) {
      results.staff = results.staff.filter((o) => staffOptions.includes(o.id));
    }

    if (!this.is_requestable && !this.is_visitable) {
      results.courses = results.courses.filter((o) => courseOptions.includes(o.id));
    }

    results.locations = results.locations.filter((o) => locationOptions.includes(o.id));
    if (results.meeting_types && Array.isArray(results.meeting_types)) {
      results.meeting_types = results.meeting_types.filter((o) => meetingTypeOptions.includes(o.id));
    }

    this.viewIndividualAvailabilitiesForLocation = this.urlParams.staff_at_location || 0;

    return results;
  }

  // Populates timesByDate on each resource
  __filterByDate(resource, date) {
    date = date.replace(/-/g, '');

    resource.timesByDate = resource.times.filter((time) => {
      return time.split('T')[0].replace(/-/g, '') == date;
    });
    return resource;
  }

  // Populates times on each resource from resources[].staff_times[].times[]
  __groupStaffTimesForLocation(resource) {
    const allTimes = resource.staff_times.reduce((tms, stf) => tms.concat(stf.times), []);
    resource.times = Array.from(new Set(allTimes)).sort();
  }

  __showResultsByDate(e) {
    const seekNext = e.detail.value === 'next';
    const seekPrevious = e.detail.value === 'previous';
    const currentSelection = this.selectedDate;
    const currentSelectionIndex = this.__availableDates.indexOf(currentSelection);
    const totalAvailableDates = this.__availableDates.length;
    if (seekPrevious && currentSelectionIndex > 0) {
      this.selectedDate = this.__availableDates[currentSelectionIndex - 1];
    }
    if (seekNext && currentSelectionIndex > -1 && currentSelectionIndex + 1 < totalAvailableDates) {
      this.selectedDate = this.__availableDates[currentSelectionIndex + 1];
    }
    if (currentSelectionIndex == -1) {
      this.selectedDate = this.__availableDates[0];
    }
    this.setURLParams({
      ...this.urlParams,
      begin_date: this.selectedDate
    });
  }

  __formatSelectedDate() {
    return formatDateToHumanReadableString(this.selectedDate);
  }

  __groupTimesByAvailableDates(allCombinedTimes = [], availableDates = []) {
    const groupedDates = {};
    availableDates.forEach((date) => {
      groupedDates[date] = allCombinedTimes.filter((time) => {
        return time.split('T')[0] == date;
      });
    });
    return groupedDates;
  }

  __isRequiredCourseSelected(appointment) {
    if (!appointment.must_pick_course) {
      return true;
    } else {
      return 'course_id' in this.urlParams;
    }
  }

  __clearAll() {
    this.getInstance('clear_all').invoke({});
  }

  // TEMPLATE PARTIALS
  __selectionTagsTemplate() {
    let selectedOptionAsTags = [];
    const urlParams = { ...this.urlParams };
    const nonRemovable = ['care_unit_id', 'service_id'];
    if (this.is_invitation) {
      nonRemovable.push('location_id');
      nonRemovable.push('course_id');
      delete urlParams.invitation_id;
    }
    if (this.is_unauthenticated) {
      nonRemovable.push('location_id');
      nonRemovable.push('course_id');
      delete urlParams.token;
    }
    delete urlParams.pal_guid;
    delete urlParams.begin_date;
    delete urlParams.staff_at_location;
    let urlParamsCopy = Object.assign({}, urlParams);
    delete urlParamsCopy.location_page;

    for (const param in urlParamsCopy) {
      const paramValues = urlParams[param];
      const availableOptionsforParam = this.filteredFormDefinition.properties[param].options;
      let selectedParamValues = Array.isArray(paramValues) ? paramValues : [paramValues];

      if (this.filteredFormDefinition.properties && availableOptionsforParam) {
        availableOptionsforParam.forEach((option) => {
          if (selectedParamValues.includes(option.id)) {
            selectedOptionAsTags.push({
              id: option.id,
              display_name: option.name
                ? option.name
                : `${option.first_name} ${option.last_name}${option.formatted_pronouns}`,
              param_name: param,
              allow_remove: !nonRemovable.includes(param)
            });
          }
        });
      }
    }

    return html`
      <div l-kind l-width-fixed class="mlm">
        <div>${this.__mobileFilterToggleTemplate()}</div>
        <div>
          <hi-tag-group class="selected_filters">
            ${repeat(
              selectedOptionAsTags,
              (tag) => `${tag.display_name}${tag.id}`,
              (tag) => {
                return html`<hi-tag
                  class="mrsx mbsx"
                  value=${tag.param_name}
                  icon=${tag.allow_remove ? '' : 'lock'}
                  theme=${tag.allow_remove ? 'primary' : 'gray-light'}
                  ?allow-remove=${tag.allow_remove}
                  @close-ended="${this.__removeFilterItem(tag.param_name, tag.id)}"
                >
                  ${tag.display_name}${tag.formatted_pronouns}</hi-tag
                >`;
              }
            )}
          </hi-tag-group>
        </div>
      </div>
    `;
  }

  __toggleMobileFilters() {
    this.mobileFiltersToggle = !this.mobileFiltersToggle;
    this.requestUpdate();
  }

  __mobileFilterToggleTemplate() {
    return html` <hi-button
      class="mobile_filters_toggle_button"
      @trigger-ended="${this.__toggleMobileFilters}"
      layout-icon
      aesthetic="ghost"
    >
      ${this.mobileFiltersToggle
        ? html`<hi-icon kind="cross" size="large-x"></hi-icon>`
        : html`<hi-icon kind="filter" size="large-x"></hi-icon>`}
    </hi-button>`;
  }

  __removeFilterItem(param, value) {
    return (event) => {
      if (Array.isArray(this.urlParams[param])) {
        this.urlParams[param] = this.urlParams[param].filter((v) => v != value);
      } else {
        delete this.urlParams[param];
      }
      this.setURLParams({ ...this.urlParams });
    };
  }

  __appointmentResults(list) {
    const noResults = list.length == 0;
    const singleLocationResult = !noResults && list.length == 1;
    const singleStaffResult = singleLocationResult && list[0].staffs && list[0].staffs.length == 1;
    const multipleResults = list.length > 0;
    // Fetching
    if (this.isFetchingResults) {
      return this.__fetchingResultsTemplate();
    }
    // Schedulable
    else if (!this.is_requestable && !this.is_visitable) {
      if (noResults) {
        return this.__emptyResultsTemplate();
      } else if (singleStaffResult) {
        return this.__singleStaffTemplate(list[0]);
      } else if (this.viewIndividualAvailabilitiesForLocation > 0) {
        const appointment = list.find((a) => a.location_id === this.viewIndividualAvailabilitiesForLocation);
        return this.__individualAvailabilitiesTemplate(appointment);
      } else if (singleLocationResult) {
        return this.__singleLocationTemplate(list[0]);
      } else {
        return this.__multipleResultsTemplate(list);
      }
    }
    // Requestable
    else if (this.is_requestable) {
      if (noResults) {
        return this.__emptyRequestableResultsTemplate();
      } else {
        return this.__requestableLocationsTemplate(list);
      }
    }
    // Visitable
    else if (this.is_visitable) {
      if (noResults && this.locationsRecord.length === 0) {
        return this.__emptyVisitableResultsTemplate();
      } else if (this.urlParams.location_id !== undefined) {
        return this.__singleVisitableLocationTemplate(list[0]);
      } else {
        return this.__multipleVisitableLocationsTemplate(list);
      }
    }
    return ``;
  }

  __locationPaginationTemplate(list) {
    const paginationContext = this.__paginationContext(list);

    return paginationContext.show_pagination
      ? html`
          <hi-pagination
            class="text-right ptmx pbmx"
            page-index="${paginationContext.list_page}"
            page-size="${paginationContext.list_per_page}"
            max="${list.length}"
            @page-index-changed=${this.gotoLocationPage()}
          >
          </hi-pagination>
          <div class="text-right">${paginationContext.pagination_text}</div>
        `
      : html``;
  }

  __paginationContext(list) {
    let paginationContext = {
      list_page: parseInt(getURLParams().location_page || 1) - 1,
      list_per_page: 25
    };

    paginationContext.show_pagination = list.length > paginationContext.list_per_page;
    paginationContext.first_index_in_range = paginationContext.list_page * paginationContext.list_per_page + 1;
    paginationContext.last_index_in_range = Math.min(
      (paginationContext.list_page + 1) * paginationContext.list_per_page,
      list.length
    );
    paginationContext.pagination_text = `Showing ${paginationContext.first_index_in_range}-${paginationContext.last_index_in_range} of ${list.length} Total Locations`;
    paginationContext.visible_list = paginationContext.show_pagination
      ? list.slice(paginationContext.first_index_in_range - 1, paginationContext.last_index_in_range)
      : list;

    return paginationContext;
  }

  __multipleResultsTemplate(list) {
    const visible_list = this.__paginationContext(list).visible_list;
    return html`
      ${this.__appointmentResultsHeaderTemplate(list.length)}
      <ul kind="no-dot">
        ${repeat(
          this.__paginationContext(list).visible_list,
          (appointment) => appointment.location.id,
          (appointment, index) =>
            this.__groupedLocationTemplate(
              appointment,
              index + 1 < visible_list.length ? visible_list[index + 1].location.id : undefined
            )
        )}
      </ul>
      ${this.__locationPaginationTemplate(list)}
    `;
  }

  __requestableLocationsTemplate(list) {
    return html`
      <h2 class="clr-gray_90-txt">${list.length} Locations</h2>
      <ul kind="no-dot">
        ${repeat(
          this.__paginationContext(list).visible_list,
          (appointment) => appointment.location.id,
          (appointment, index) => this.__requestableLocationTemplate(appointment)
        )}
      </ul>
      ${this.__locationPaginationTemplate(list)}
    `;
  }

  __requestableLocationTemplate(appointment) {
    return html`
      <li class="result-item-card clr-white-bkg pamx bbs bdr-solid clr-gray_30-bdr">
        <p class="text-sz-lx text-normal">${appointment.location.name}</p>
        <div class="location_description">
          ${appointment.location.description
            ? html`<p class="mtsxx">
                ${unsafeHTML(appointment.location.description?.replace(/href=/g, 'target="_blank" href='))}
              </p>`
            : ''}
        </div>
        ${this.__isRequiredCourseSelected(appointment)
          ? html`<hi-button class="mtm" aesthetic="secondary" @trigger-ended="${this.requestTimeCallback(appointment)}"
              >Request Time</hi-button
            >`
          : html`<div class="mtm">${this.__requestableCourseRequiredTemplate(this.courseRecord)}</div>`}
      </li>
    `;
  }

  __singleVisitableLocationTemplate(visitableResults) {
    const isRequiredCourseSelected = this.__isRequiredCourseSelected(visitableResults);
    let courseIds = visitableResults.dropin_times.reduce((arr, st) => {
      return arr.concat(st.course_ids);
    }, []);
    courseIds = Array.from(new Set(courseIds));

    return html`
      <h2>${visitableResults.location.name}</h2>
      <div class="location_description">
        ${visitableResults.location.description
          ? html`<p class="mtsxx">
              ${unsafeHTML(visitableResults.location.description?.replace(/href=/g, 'target="_blank" href='))}
            </p>`
          : ''}
      </div>
      ${visitableResults.currently_available_for_first_available_dropins
        ? html`<div class="ptm pbm bbs bdr-solid clr-gray_30-bdr">
            ${isRequiredCourseSelected
              ? html`<hi-button aesthetic="secondary" @trigger-ended="${this.virtualCheckin()}"
                  >Check-in with first available</hi-button
                >`
              : this.__visitableCourseRequiredTemplate(courseIds)}
          </div>`
        : ''}
      <ul kind="no-dot">
        ${repeat(visitableResults.dropin_times, (dropin_time) =>
          this.__dropinAvailabilityTemplate(dropin_time, isRequiredCourseSelected)
        )}
      </ul>
    `;
  }

  __dropinAvailabilityTemplate(dropin_time, isRequiredCourseSelected) {
    return html`
      <li class="result-item-card clr-white-bkg pamx bbs bdr-solid clr-gray_30-bdr">
        <p class="text-sz-lx text-bold">
          ${dropin_time.staff.first_name} ${dropin_time.staff.last_name}${dropin_time.staff.formatted_pronouns}
        </p>
        ${dropin_time.staff.relationships.length > 0
          ? html`<p>
              ${' (Your ' + dropin_time.staff.relationships.join(', ').replace(/,\s([^,]+)$/, ' and $1') + ')'}
            </p>`
          : ``}
        <div class="plm mtsxx">
          ${dropin_time.times.map((time) => {
            let detailsContent = html`
              <div class="availability_details text-light clr-gray_20-bkg ptsx plsx">${unsafeHTML(time.details)}</div>
            `;

            let meetingTypesContent = html` <p class="text-sz-l">Meeting Types: ${time.meeting_types}</p> `;
            return html`
              <div class="mbm">
                <p class="text-sz-l">${time.weekdays}</p>
                <p class="text-sz-l">${formatRangeTime(time.begin_time, time.end_time)}</p>
                <p class="text-sz-l">${time.range}</p>
                ${time.meeting_types ? meetingTypesContent : ''} ${time.details !== '' ? detailsContent : ''}
              </div>
            `;
          })}
          ${dropin_time.currently_available_for_dropins && isRequiredCourseSelected
            ? html`<hi-button aesthetic="secondary" @trigger-ended="${this.virtualCheckin(dropin_time.staff.id)}"
                >Drop-in Online</hi-button
              >`
            : dropin_time.currently_available_for_dropins && !isRequiredCourseSelected
            ? this.__visitableCourseRequiredTemplate(dropin_time.course_ids)
            : ``}
        </div>
      </li>
    `;
  }

  __multipleVisitableLocationsTemplate(list) {
    return html`
      <h2 class="clr-gray_90-txt">${list.length} Locations</h2>
      <ul kind="no-dot">
        ${repeat(
          this.__paginationContext(list).visible_list,
          (resource) => resource.location.id,
          (resource, index) => this.__visitableLocationsTemplate(resource.location)
        )}
      </ul>
      ${this.__locationPaginationTemplate(list)}
    `;
  }

  __visitableLocationsTemplate(location) {
    return html`
      <li class="result-item-card clr-white-bkg pamx bbs bdr-solid clr-gray_30-bdr">
        <div>
          <hi-link href="#" @trigger-ended="${this.gotoLocationCallback(location.id)}">
            <p class="text-sz-lx text-bold">${location.name}</p>
          </hi-link>
          <div class="location_description">
            ${location.description
              ? html`<p class="mtsxx">
                  ${unsafeHTML(location.description?.replace(/href=/g, 'target="_blank" href='))}
                </p>`
              : ''}
          </div>
        </div>
      </li>
    `;
  }

  __dateWidgetTemplate() {
    return html`
      <div class="bbs bdr-solid bts clr-gray_20-bdr">
        <div class="maa max-w-240" l-kind="hort" l-width-fixed l-align="center" l-justify="between">
          <hi-button
            aesthetic="ghost"
            class="pam baz max-w-40"
            value="previous"
            @trigger-ended="${this.__showResultsByDate}"
          >
            <hi-icon kind="chevron_left"></hi-icon>
            <span class="hide-ax">Previous</span>
          </hi-button>
          <p class="pam text-center" l-size="grow">${this.__formatSelectedDate()}</p>
          <hi-button
            aesthetic="ghost"
            class="pam baz max-w-40"
            value="next"
            @trigger-ended="${this.__showResultsByDate}"
          >
            <span class="hide-ax">Next</span>
            <hi-icon kind="chevron_right"></hi-icon>
          </hi-button>
        </div>
      </div>
    `;
  }

  __appointmentResultsHeaderTemplate(count) {
    return html`
      <h2 class="clr-gray_90-txt">${count} Locations</h2>
      ${this.__dateWidgetTemplate()}
    `;
  }

  __locationDropinTemplate(appointment) {
    return appointment.dropin_available
      ? html`<hi-link href="#" size="small" @trigger-ended="${this.viewAvailableDropinTimes(appointment)}"
          >Drop-in Times Available</hi-link
        >`
      : ``;
  }

  __groupedLocationTemplate(appointment, nextLocationId) {
    let courseIds = appointment.staff_times.reduce((arr, st) => {
      return arr.concat(st.course_ids);
    }, []);
    courseIds = Array.from(new Set(courseIds));

    return html`
      <li class="result-item-card clr-white-bkg pamx bbs bdr-solid clr-gray_30-bdr">
        <div class="mbsxx">
          <hi-link
            .id=${`location-${appointment.location_id}`}
            href="#"
            @trigger-ended="${this.gotoLocationCallback(appointment.location_id)}"
          >
            <p class="text-sz-lx text-bold">${appointment.location.name}</p>
          </hi-link>
          ${nextLocationId
            ? html` <a
                class="visually-hidden mlm"
                href="#"
                @trigger-ended=${(e) => {
                  e.preventDefault();
                }}
                @keyup=${(e) => {
                  if (e.key === 'Enter') {
                    document.getElementById(`location-${nextLocationId}`).focus();
                  }
                }}
                @focusin=${(e) => {
                  e.preventDefault();
                  e.target.classList.remove('visually-hidden');
                }}
                @blur=${(e) => {
                  e.target.classList.add('visually-hidden');
                }}
                >Skip to next available Location</a
              >`
            : ''}
          <div class="location_description">
            ${appointment.location.description
              ? html`<p class=" mtsxx">
                  ${unsafeHTML(appointment.location.description?.replace(/href=/g, 'target="_blank" href='))}
                </p>`
              : ''}
          </div>
        </div>
        ${this.__locationDropinTemplate(appointment)}
        ${appointment.can_choose_staff === undefined || appointment.can_choose_staff
          ? this.__avatarsTemplate(appointment)
          : ``}
        <div class="ptm">
          ${this.__isRequiredCourseSelected(appointment)
            ? timesTemplate(appointment.timesByDate, appointment.duration, this.selectTimeCallback(appointment))
            : this.__schedulableCourseRequiredTemplate(courseIds)}
        </div>
      </li>
    `;
  }

  __emptyResultsTemplate() {
    const message =
      this.is_invitation || this.is_unauthenticated
        ? 'There are no more available appointments at this time.'
        : 'Update your selected options to see more availabilities or Request alternate time';
    return html`
      <div class="text-center pamx">
        <p class="text-sz-lx text-bold">No Appointments available!</p>
        <p class="clr-gray_70-txt">${message}</p>
      </div>
    `;
  }

  __emptyRequestableResultsTemplate() {
    return html`
      <div class="text-center pamx">
        <p class="text-sz-lx text-bold">No locations accepting requests for alternate times!</p>
        <p class="clr-gray_70-txt"></p>
      </div>
    `;
  }

  __emptyVisitableResultsTemplate() {
    return html`
      <div class="text-center pamx">
        <p class="text-sz-lx text-bold">No drop-in times available!</p>
        <p class="clr-gray_70-txt"></p>
      </div>
    `;
  }

  __fetchingResultsTemplate() {
    return html`<div class="text-center pamx min-h-360">
      <div l-kind l-width-fixed l-align="center" l-justify="center" class="min-h-120">
        <hi-indicator></hi-indicator>
      </div>
      <p class="text-sz-lx text-bold">
        Loading ${this.is_requestable || this.is_visitable ? 'Locations' : 'Appointments'}...
      </p>
      <p class="clr-gray_70-txt"></p>
    </div>`;
  }

  __dropinLinkTemplate(appointment, params) {
    return appointment.dropin_available
      ? html`<hi-link href="#" @trigger-ended="${this.viewAvailableDropinTimes(appointment, params)}"
          >View Drop-in Times</hi-link
        >`
      : ``;
  }

  __singleStaffTemplate(appointment) {
    const selectedStaff = appointment.staffs[0];
    const selectedStaffTimes = appointment.staff_times[0].times;
    return html`
      <div class="clr-white-bkg bbs bdr-solid clr-gray_30-bdr">
        ${appointment.can_choose_staff === undefined || appointment.can_choose_staff
          ? html` <div class="mbsx" l-kind l-width-fixed>
              <div>${this.__staffAvatarTemplate(selectedStaff, 'large')}</div>
              <div class="mlm">
                <h2>${selectedStaff.first_name} ${selectedStaff.last_name}${selectedStaff.formatted_pronouns}</h2>
                <p class="text-sz-lx clr-gray_80-txt">${appointment.location.name}</p>
                <div class="location_description">
                  ${appointment.location.description
                    ? html`<p class="mtsxx mbsxx">
                        ${unsafeHTML(appointment.location.description?.replace(/href=/g, 'target="_blank" href='))}
                      </p>`
                    : ''}
                </div>
                ${this.__dropinLinkTemplate(appointment, { staff_ids: [appointment.staffs[0].id] })}
              </div>
            </div>`
          : html`
              <div>
                <h2>${appointment.location.name}</h2>
                <div class="location_description">
                  ${appointment.location.description
                    ? html`<p class="mtsxx mbsxx">
                        ${unsafeHTML(appointment.location.description?.replace(/href=/g, 'target="_blank" href='))}
                      </p>`
                    : ''}
                </div>
                ${this.__dropinLinkTemplate(appointment, {})}
              </div>
            `}
        ${this.__isRequiredCourseSelected(appointment)
          ? this.__groupedTimesByDayTemplate(selectedStaffTimes, appointment)
          : html`<div class="mtm mbm">
              ${this.__schedulableCourseRequiredTemplate(appointment.staff_times[0].course_ids)}
            </div>`}
      </div>
    `;
  }

  __singleLocationTemplate(appointment) {
    let courseIds = appointment.staff_times.reduce((arr, st) => {
      return arr.concat(st.course_ids);
    }, []);
    courseIds = Array.from(new Set(courseIds));

    return html`
      <div class="clr-white-bkg bbs bdr-solid clr-gray_30-bdr">
        <div class="mbmsx">
          <h2>${appointment.location.name}</h2>
        </div>
        <div class="location_description">
          ${appointment.location.description
            ? html`<p class="mtsxx mbsxx">
                ${unsafeHTML(appointment.location.description?.replace(/href=/g, 'target="_blank" href='))}
              </p>`
            : ''}
        </div>
        ${this.__dropinLinkTemplate(appointment)}
        ${appointment.can_choose_staff === undefined || appointment.can_choose_staff
          ? html`${this.__avatarsTemplate(appointment)}
              <hi-link href="#" @trigger-ended="${this.showIndividualAvailabilities(appointment.location_id)}"
                >View individual availabilities</hi-link
              >`
          : ``}
        ${this.__isRequiredCourseSelected(appointment)
          ? this.__groupedTimesByDayTemplate(appointment.times, appointment)
          : html`<div class="mtm mbm">${this.__schedulableCourseRequiredTemplate(courseIds)}</div>`}
      </div>
    `;
  }

  __groupedTimesByDayTemplate(times, appointment) {
    //TODO: use repeat directive to enhance performance?
    const groupedTimes = this.__groupTimesByAvailableDates(times, this.__availableDates);
    const timesListTemplates = [];
    const groupedTimesArr = Object.keys(groupedTimes);
    groupedTimesArr.forEach((day, index) => {
      timesListTemplates.push(html`
        <li class="mbm">
          <div class="pasx mbm plm clr-gray_20-bkg">
            <p .id=${`skip-${day}`} tabindex="0" class="display-inline ${day === this.selectedDate ? 'text-bold' : ''}">
              ${formatDateToHumanReadableString(day)}
            </p>
            ${index + 1 < groupedTimesArr.length
              ? html` <a
                  .id=${`skip-${day}-link`}
                  class="visually-hidden mlm"
                  href="#"
                  @trigger-ended=${(e) => {
                    e.preventDefault();
                    document.getElementById(`skip-${groupedTimesArr[index + 1]}`).focus();
                  }}
                  @focusin=${(e) => {
                    e.preventDefault();
                    e.target.classList.remove('visually-hidden');
                  }}
                  @focus=${(e) => {
                    e.preventDefault();
                  }}
                  @blur=${(e) => {
                    e.target.classList.add('visually-hidden');
                  }}
                  >Skip to next available day ${formatDateToHumanReadableString(groupedTimesArr[index + 1])}</a
                >`
              : ''}
          </div>
          <div>${timesTemplate(groupedTimes[day], appointment.duration, this.selectTimeCallback(appointment))}</div>
        </li>
      `);
    });
    return html` <ul class="ptm" kind="no-dot">
      ${timesListTemplates}
    </ul>`;
  }

  __individualAvailabilitiesTemplate(location) {
    const people = location.staffs;
    return html`
      <ul kind="no-dot">
        ${repeat(
          people,
          (person) => person.id,
          (person) => html`${this.__personTemplate(person)}`
        )}
      </ul>
    `;
  }

  __personTemplate(person) {
    return html`
      <li
        class="result-item-card pam bdr-solid bbs clr-gray_40-bdr"
        l-kind="hort"
        l-justify="between"
        l-width-fixed
        l-align="center"
      >
        <div class="display-flex">
          <div>
            <hi-avatar initials="${person.first_name[0] + person.last_name[0]}" image="${person.picture}"></hi-avatar>
          </div>
          <div class="mll">
            <hi-link href="#" @trigger-ended="${this.gotoPersonCallback(person.id)}">
              <p class="text-sz-lx pbsx">${person.first_name} ${person.last_name}${person.formatted_pronouns}</p>
            </hi-link>
            <p>${person.relationships.join(', ')}</p>
          </div>
        </div>
        <div>
          <hi-icon kind="chevron_right" theme="gray-dark" size="large-x"></hi-icon>
        </div>
      </li>
    `;
  }

  __schedulableCourseRequiredTemplate(courseIds) {
    return this.__courseRequiredTemplate(
      'Please select one of the courses below to see available times at this location.',
      courseIds
    );
  }

  __requestableCourseRequiredTemplate(courses) {
    return this.__courseRequiredTemplate(
      'Please select one of the courses below before requesting a time at this location.',
      courses
    );
  }

  __visitableCourseRequiredTemplate(courseIds) {
    return this.__courseRequiredTemplate('Please select one of the courses below before checking in.', courseIds);
  }

  __courseRequiredTemplate(message, courses) {
    return html`
      <div class="pam bdr-round-m clr-cyan_10-bkg clr-cyan_40-bdr bas bdr-solid">
        <div class="display-flex">
          <hi-icon kind="info_circle_filled" theme="blue"></hi-icon>
          <p class="mlsxxx text-info">${message}</p>
        </div>
        <ul class="text-info text-sz-m mllxxx">
          ${courses.map((course) => {
            if (tt.is('number', course)) {
              course = this.courseRecord.find((c) => c.id === course);
            }
            return html`<li>
              <hi-link href="#" @trigger-ended=${this.gotoCourseCallback(course.id)}
                >${course.name} (${course.number})</hi-link
              >
            </li>`;
          })}
        </ul>
      </div>
    `;
  }

  __escapeHatchTemplate(schedulableHatch, requestableHatch, visitableHatch, palHatch) {
    if (schedulableHatch != '' || requestableHatch != '' || visitableHatch != '' || palHatch != '') {
      return html`<div class="ptmx">
        <span class="text-sz-lx">Don’t see anything that works for you?</span>
        ${schedulableHatch} ${requestableHatch} ${visitableHatch} ${palHatch}
      </div>`;
    }

    return '';
  }

  __palEscapeTemplate() {
    if (this.urlParams.hasOwnProperty('pal_guid')) {
      return html`<div class="bas bdr-solid clr-gray_30-bdr pam masm mtm bdr-round-m">
        <hi-link href="#/my/appointment-new" @trigger-ended="${this.clearAll}">
          <span class="text-sz-lx">See All Scheduling Options</span>
        </hi-link>
        <p>
          You are currently viewing options that are limited to the availabilities in the link you clicked.
          <hi-link href="#/my/appointment-new"> Click here </hi-link> to go back to main scheduling workflow for all
          other options.
        </p>
      </div>`;
    }

    return '';
  }
  __seeSchedulableTimesTemplate() {
    const baseSchedulableData = this.getDefinition('base_schedulable_data');
    if (this.is_requestable || this.is_visitable) {
      if (baseSchedulableData && baseSchedulableData.length > 0) {
        const serviceId = this.urlParams.service_id;
        const serviceIsIncluded = baseSchedulableData.some(function (s) {
          return s === serviceId;
        });

        if (serviceIsIncluded) {
          return html`<div class="bas bdr-solid clr-gray_30-bdr pam masm mtm bdr-round-m">
            <hi-link href="#" @trigger-ended="${this.scheduleApptTimeAction}">
              <span class="text-sz-lx">Schedule an Appointment</span>
            </hi-link>
            <p>There may be other locations that support scheduling Appointments.</p>
          </div>`;
        }
      }
    }

    return '';
  }

  __requestAlternateTimeTemplate() {
    if (this.urlParams.hasOwnProperty('pal_guid')) {
      return '';
    }

    const baseRequestableData = this.getDefinition('base_requestable_data');
    if (!this.is_requestable && !this.is_invitation) {
      if (baseRequestableData && baseRequestableData.length > 0) {
        const serviceId = this.urlParams.service_id;
        const serviceIsIncluded = baseRequestableData.some(function (s) {
          return s === serviceId;
        });

        if (serviceIsIncluded) {
          return html`<div class="bas bdr-solid clr-gray_30-bdr pam masm mtm bdr-round-m">
            <hi-link href="#" @trigger-ended="${this.requestAlternateTimeAction}">
              <span class="text-sz-lx">Request Alternate Appointment Time</span>
            </hi-link>
            <p>There may be other locations that support Appointment Requests.</p>
          </div>`;
        }
      }
    }

    return '';
  }

  __seeOtherDropinTimesTemplate() {
    if (this.urlParams.hasOwnProperty('pal_guid')) {
      return '';
    }

    const baseVisitableData = this.getDefinition('base_visitable_data');
    if (!this.is_requestable && !this.is_invitation && !this.is_visitable) {
      if (baseVisitableData && baseVisitableData.length > 0) {
        const serviceId = this.urlParams.service_id;
        const serviceIsIncluded = baseVisitableData.some(function (s) {
          return s === serviceId;
        });

        if (serviceIsIncluded) {
          return html`<div class="bas bdr-solid clr-gray_30-bdr pam masm mtm bdr-round-m">
            <hi-link href="#" @trigger-ended="${this.seeOtherDropinTimes}">
              <span class="text-sz-lx">View Drop-in Times</span>
            </hi-link>
            <p>There may be other locations that support Drop-ins.</p>
          </div>`;
        }
      }
    }

    return '';
  }

  __alertTemplate() {
    if (this.results && this.results.campaign && this.results.campaign.note != '') {
      return html`
        <hi-alert theme="warning" class="mbm" mode="presentation">
          <p>${this.results.campaign.note}</p>
        </hi-alert>
      `;
    }
  }

  __avatarsTemplate(appointment) {
    const staffs = appointment.staffs;
    const personWord = staffs.length === 1 ? 'Person' : 'People';

    if (staffs.length > 15) {
      return html`<div l-kind l-width-fixed l-align="center" class="ptm">
        <hi-link
          href="#"
          @trigger-ended="${this.showIndividualAvailabilities(appointment.location_id)}"
          class="text-sz-m"
          >${staffs.length} ${personWord}</hi-link
        >
      </div>`;
    }

    return html`<div l-kind l-width-fixed l-align="center" class="ptm">
      <hi-avatar-group> ${repeat(staffs, this.__staffAvatarTemplate)} </hi-avatar-group>
      <span class="text-sz-m mll">
        <hi-link
          href="#"
          @trigger-ended="${this.showIndividualAvailabilities(appointment.location_id)}"
          class="text-sz-m"
          >${staffs.length} ${personWord}</hi-link
        >
      </span>
    </div>`;
  }

  __staffAvatarTemplate(staff, size = 'small') {
    return html` <hi-avatar
      class="non-interactive"
      size="${size}"
      initials="${staff.first_name[0] + staff.last_name[0]}"
      image="${staff.picture}"
    ></hi-avatar>`;
  }

  // ACTIONS

  selectTimeCallback(appointment) {
    return (event) => {
      const selectAppointmentAction = this.getInstance('select_appointment');
      const def = this.getDefinition('filters_type');
      const results = this.getDefinition('results');
      const {
        // Filter away the staff_times property
        // because it is unused and potentially
        // GINORMOUS.
        staff_times,
        ...smallAppointmentObject
      } = appointment;
      const selected_appointment_values = {
        begin_time: event.detail.value,
        duration: appointment.duration,
        service_id: appointment.service_id,
        location_id: appointment.location_id,
        care_unit_id: appointment.care_unit_id,
        course_id: this.urlParams.course_id,
        pal_guid: this.urlParams.pal_guid,
        invitation_id: this.invitationId,
        token: this.token,
        appointment: smallAppointmentObject,
        courses: def.properties.course_id.options,
        communication_settings: results.communication_settings
      };
      if (def.properties.meeting_type_id) {
        selected_appointment_values.meeting_types = def.properties.meeting_type_id.options;

        if (this.urlParams.meeting_type_id !== undefined) {
          selected_appointment_values.meeting_type_id = this.urlParams.meeting_type_id;
        }
      }
      if (this.urlParams.staff_ids !== undefined) {
        selected_appointment_values.staff_id = this.urlParams.staff_ids;
      } else if (appointment.staffs.length === 1) {
        selected_appointment_values.staff_id = appointment.staffs[0].id;
      }
      if (selectAppointmentAction) {
        selectAppointmentAction.invoke(selected_appointment_values);
      }
    };
  }

  requestTimeCallback(appointment) {
    return (event) => {
      const selectAppointmentAction = this.getInstance('select_appointment');
      const def = this.getDefinition('filters_type');
      const results = this.getDefinition('results');
      const selected_appointment_values = {
        service_id: appointment.service_id,
        location_id: appointment.location_id,
        care_unit_id: appointment.care_unit_id,
        course_id: this.urlParams.course_id,
        appointment: appointment,
        courses: def.properties.course_id.options
      };
      if (selectAppointmentAction) {
        selectAppointmentAction.invoke(selected_appointment_values);
      }
    };
  }

  virtualCheckin(staffId) {
    return (event) => {
      const virtualCheckinAction = this.getInstance('virtual_checkin');
      const params = {
        care_unit_id: parseInt(this.urlParams.care_unit_id, 10),
        service_id: parseInt(this.urlParams.service_id, 10),
        location_id: parseInt(this.urlParams.location_id, 10),
        course_id: parseInt(this.urlParams.course_id, 10),
        staff_id: staffId
      };
      if (virtualCheckinAction) {
        virtualCheckinAction.invoke(params);
      }
    };
  }

  gotoLocationPage() {
    return (event) => {
      let pageIndex = event.detail.value;
      pageIndex = pageIndex + 1; // pagination API is 0-indexed, but we want our URL to be 1-indexed

      if (pageIndex == 1) {
        delete this.urlParams.location_page;
      } else {
        this.urlParams = { ...this.urlParams, location_page: pageIndex };
      }

      this.setURLParams(this.urlParams);
    };
  }

  gotoLocationCallback(locationId) {
    return (event) => {
      this.urlParams = { ...this.urlParams, location_id: locationId };

      this.setURLParams(this.urlParams);
      if (this.is_visitable) {
        this.getResults(this.urlParams);
      }
    };
  }

  gotoPersonCallback(personId) {
    return (event) => {
      this.urlParams = { ...this.urlParams, staff_ids: personId };
      delete this.urlParams.staff_at_location;
      this.viewIndividualAvailabilitiesForLocation = 0;
      this.setURLParams(this.urlParams);
    };
  }

  gotoCourseCallback(courseId) {
    return (event) => {
      this.setURLParams({ ...this.urlParams, course_id: courseId });
    };
  }

  showIndividualAvailabilities(locationId) {
    return (event) => {
      this.urlParams = { ...this.urlParams, staff_at_location: locationId };
      delete this.urlParams.location_page;
      this.setURLParams(this.urlParams);
    };
  }

  seeOtherDropinTimes() {
    const seeDropinsAction = this.getInstance('view_available_dropins');
    seeDropinsAction.invoke({
      care_unit_id: this.urlParams.care_unit_id,
      service_id: this.urlParams.service_id,
      begin_date: this.__getCurrentDateFormatted
    });
  }

  viewAvailableDropinTimes(appointment, params = {}) {
    const availableDropins = this.getInstance('view_available_dropins');

    return () => {
      const locationParams = {
        location_id: appointment.location.id
      };

      let urlParamsCopy = Object.assign({}, this.urlParams);

      delete urlParamsCopy.begin_date;
      delete urlParamsCopy.location_page;

      availableDropins.invoke({
        ...urlParamsCopy,
        ...locationParams,
        ...params
      });
    };
  }

  scheduleApptTimeAction() {
    const scheduleApptTimeAction = this.getInstance('appointment_results');
    scheduleApptTimeAction.invoke({
      care_unit_id: this.urlParams.care_unit_id,
      service_id: this.urlParams.service_id,
      begin_date: this.__getCurrentDateFormatted
    });
  }

  requestAlternateTimeAction() {
    const requestApptAction = this.getInstance('request_alternate');
    requestApptAction.invoke({
      care_unit_id: this.urlParams.care_unit_id,
      service_id: this.urlParams.service_id
    });
  }

  getVisitables() {
    const getVisitablesAction = this.getInstance('get_visitables');
    if (getVisitablesAction) {
      getVisitablesAction.invoke();
    }
  }

  getRequestables() {
    const getRequestablesAction = this.getInstance('get_requestables');
    if (getRequestablesAction) {
      getRequestablesAction.invoke();
    }
  }

  getSchedulables() {
    const getSchedulablesAction = this.getInstance('get_schedulables');
    if (getSchedulablesAction) {
      getSchedulablesAction.invoke();
    }
  }

  setURLParams(data) {
    if (this.is_visitable) {
      delete data['begin_date'];
    }

    const { staff_ids, course_id, meeting_type_id } = getURLParams();

    if (
      (data.begin_date && Object.keys(this.datesRequested).length > 0 && !this.datesRequested[data.begin_date]) ||
      (data.staff_ids || '') != (staff_ids || '') ||
      (data.course_id || '') != (course_id || '') ||
      (data.meeting_type_id || '') != (meeting_type_id || '') ||
      data.location_id
    ) {
      delete data.location_page;
    }

    for (let prop in data) {
      if (data[prop] === '' || data[prop] === undefined || data[prop] === null) {
        delete data[prop];
      }
    }
    // Can have URL that looks like: /capabilities?invitation_token=vUewjcRZsO#/my/appointment-invitation-claim-success?token=vUewjcRZsO
    const serializedQueryParams = window.safe.serializeParameters(data);
    const [urlBeforeHash, urlAfterHash] = window.location.href.split('#');
    const [urlPathName, currentParams] = urlAfterHash.split('?');
    const urlPartWithoutParams = `${urlBeforeHash}#${urlPathName}`;
    //TODO: Only update hash if the params are different. Prevents a second loop & update when params are number vs string. E.g. staff_ids: 1234 vs staff_ids: "1234"
    if (serializedQueryParams != currentParams) {
      window.location.assign(`${urlPartWithoutParams}?${serializedQueryParams}`);
    }
  }
}

// customElements.define('ai-prefab-appointment-results', PrefabAppointmentResultsElement);
HIP.factory.element(PrefabAppointmentResultsElement);
