import { LitElement, html } from 'https://cdn.eab.com/toolkit/21-4-12/hip/library/lit-element/lit-element.js';
import { unsafeHTML } from 'https://cdn.eab.com/toolkit/21-4-12/hip/library/lit-html/directives/unsafe-html.js';

class PrefabConfirmAppointmentRequestElement extends HIP.BehaviorPrefab(HIP.BehaviorElement(LitElement)) {
  static get is() {
    return 'ai-prefab-confirm-appointment-request';
  }

  constructor() {
    super();
    this.filteredFormDefinition = {};
    this.student_scheduler_button_text = 'Request';
    this.isSubmitting = false;
  }

  connectedCallback() {
    super.connectedCallback();

    const contextOutput = this.getInstance('context').output;

    this.getInstance('initialize').addEventListener('invoke-ended', () => {});

    this.getInstance('selected_appointment').addEventListener('output-changed', (selectedAppointment) => {
      if (!this.selectedAppointment) {
        this.selectedAppointment = selectedAppointment.detail.value;

        setTimeout(() => {
          this.updateFilterDef();
          this.requestUpdate();
        }, 0);
      }
    });

    this.getInstance('confirm_appointment').addEventListener('output-changed', () => {
      this.isSubmitting = false;
    });
  }

  disconnectedCallback() {
    super.disconnectedCallback();
  }

  static get properties() {
    return {
      /**
       * all links are direct and not altered to work as relative links (client-side)
       */
      isDirect: Boolean,
      /**
       * the navigation bucket
       */
      index: Object
    };
  }

  get defaultScript() {
    return {
      system: {
        state: 'main'
      },
      action: {},
      bucket: {
        context: {},
        record: {}
      },
      io: {}
    };
  }

  getURLParams() {
    const searchParams =
      window.location.hash.split('?').length > 1 ? `?${window.location.hash.split('?')[1]}` : window.location.search;
    return new URLSearchParams(searchParams);
  }

  render() {
    const customContentText = this.getDefinition('custom_content');

    if (!this.selectedAppointment || !customContentText) {
      return this.__fetchingResultsTemplate();
    } else {
      return html`<ai-screen gid="_screen" ns="${this.ns}">
        <ai-state gid="_state__main" ns="${this.ns}" states="main" class="">
          <div l-kind="hort">
            <div class="select_care_unit_service mbl" l-size="12">
              ${this.__stylesTemplate()}
              <div class="appointment_details clr-gray_30-bdr bdr-round-all bdr-solid bas pam mbm">
                ${this.__detailsTemplate()}
              </div>
              <div class="form_container clr-gray_30-bdr bdr-round-all bdr-solid bas plm prm">
                ${this.__formTemplate()}
              </div>
            </div>
          </div>
        </ai-state>
      </ai-screen> `;
    }
  }

  getInstance(instanceName) {
    return HIP.db.gid.queryOne(instanceName, this.ns);
  }

  getDefinition(bucketName) {
    const bucket = this.getInstance(bucketName);
    if (bucket) {
      return bucket.output;
    }
  }

  // OPERATING LOGIC

  __submitForm(event) {
    if (this.isSubmitting) return;

    if (event.detail.value !== undefined) {
      this.isSubmitting = true;
      this.getInstance('submit_record').invoke(event.detail.value);
    }
  }

  updateFilterDef(formData) {
    const def = tt.deepClone(this.getDefinition('record_type') || {});
    const customContentText = this.getDefinition('custom_content') || {};

    /**
     * Below code is mapping dropdown values & setting picked options from URL.
     */
    if (def) {
      // Udpate custom labels
      if (customContentText.student_scheduling.care_unit_label) {
        this.care_unit_label = customContentText.student_scheduling.care_unit_label;
      }
      if (customContentText.student_scheduling.service_label) {
        this.service_label = customContentText.student_scheduling.service_label;
      }
      if (customContentText.student_scheduling.location_label) {
        this.location_label = customContentText.student_scheduling.location_label;
      }

      this.filteredFormDefinition = def;
    }
    this.requestUpdate();
  }

  // TEMPLATE PARTIALS

  __fetchingResultsTemplate() {
    return html`<div class="text-center pamx min-h-360">
      <div l-kind l-width-fixed l-align="center" l-justify="center" class="min-h-120">
        <hi-indicator></hi-indicator>
      </div>
      <p class="text-sz-lx text-bold">Loading Appointment...</p>
      <p class="clr-gray_70-txt"></p>
    </div>`;
  }

  __stylesTemplate() {
    return html`
      <style>
        dt {
          font-weight: var(--hi-font--weight-normal);
          font-size: var(--hi-font-size--normal);
          padding-bottom: var(--hi-size--small-x);
        }
        dl,
        dd {
          font-size: var(--hi-font-size--normal);
        }
        dl > div {
          border-bottom: 1px solid var(--hi-color--gray-30);
          padding-top: var(--hi-size--medium);
          padding-bottom: var(--hi-size--medium);
        }
        dl > div:last-child {
          border-bottom: none;
        }
        dd {
          color: var(--hi-color--gray-80);
          font-weight: var(--hi-font--weight-light);
        }
        dd > div {
          line-height: 1.5rem;
        }
      </style>
    `;
  }

  __courseInfoTemplate() {
    if (this.selectedAppointment.course_id) {
      const course_id = this.selectedAppointment.course_id;
      const course = this.selectedAppointment.courses.find((c) => c.id == course_id);
      return html`
        <div l-size="12" l-sml-size="6">
          <dt>Course</dt>
          <dd>${course.title}</dd>
        </div>
      `;
    }

    return ``;
  }

  __detailsTemplate() {
    return html`
      <dl l-kind="hort" l-justify="left">
        <div l-size="12" l-sml-size="6">
          <dt>${this.care_unit_label}</dt>
          <dt></dt>
          <dd>${this.selectedAppointment.appointment.careUnit.title}</dd>
        </div>
        <div l-size="12" l-sml-size="6">
          <dt>${this.service_label}</dt>
          <dd>${this.selectedAppointment.appointment.service.title}</dd>
        </div>
        <div l-size="12" l-sml-size="6">
          <dt>${this.location_label}</dt>
          <dd>
            <div>${this.selectedAppointment.appointment.location.title}</div>
            <div class="location_description">
              ${this.selectedAppointment.appointment.location.description
                ? html`${unsafeHTML(
                    this.selectedAppointment.appointment.location.description?.replace(
                      /href=/g,
                      'target="_blank" href='
                    )
                  )}`
                : ''}
            </div>
          </dd>
        </div>
        ${this.__courseInfoTemplate()}
      </dl>
    `;
  }

  __formTemplate() {
    return html`
      <hi-form
        ns="${this.ns}"
        id="_filters_form"
        gid="_filters_form"
        aesthetic="invisible"
        flush
        has-controls
        disable-http=""
        @submit-ended="${this.__submitForm}"
      >
        <ai-builder-form gid="_filters_form_builder" ns="${this.ns}" .input=${this.filteredFormDefinition}>
        </ai-builder-form>
        <hi-button aesthetic="primary" slot="controls" type="submit">Request</hi-button>
      </hi-form>
    `;
  }
}

// customElements.define('ai-prefab-confirm-appointment', PrefabConfirmAppointmentElement);
HIP.factory.element(PrefabConfirmAppointmentRequestElement);
