import { LitElement, html } from 'https://cdn.eab.com/toolkit/21-4-12/hip/library/lit-element/lit-element.js';
import { unsafeHTML } from 'https://cdn.eab.com/toolkit/21-4-12/hip/library/lit-html/directives/unsafe-html.js';

class PrefabConfirmAppointmentElement extends HIP.BehaviorPrefab(HIP.BehaviorElement(LitElement)) {
  static get is() {
    return 'ai-prefab-confirm-appointment';
  }

  constructor() {
    super();
    this.filteredFormDefinition = {};
    this.student_scheduler_button_text = 'Schedule';
    this.isSubmitting = false;
  }

  connectedCallback() {
    super.connectedCallback();

    const contextOutput = this.getInstance('context').output;
    this.is_invitation = !!contextOutput.is_invitation;
    this.is_unauthenticated = !!contextOutput.is_unauthenticated;

    this.token = this.getToken();

    this.getInstance('initialize').addEventListener('invoke-ended', () => {});

    this.getInstance('all_appointments').addEventListener('output-changed', () => {
      this.updateFilterDef();
    });

    this.getInstance('selected_appointment').addEventListener('output-changed', (selectedAppointment) => {
      if (!this.selectedAppointment) {
        this.selectedAppointment = selectedAppointment.detail.value;
        this.updateFilterDef();
        this.getAllAppointmentsAction();
        this.getDetailsAction(this.selectedAppointment);
      }
    });

    this.getInstance('details').addEventListener('output-changed', (details) => {
      this.details = details.detail.value;

      if (this.selectedAppointment.appointment.staffs) {
        if (this.selectedAppointment.appointment.staffs.length === 1) {
          this.staff = this.selectedAppointment.appointment.staffs[0];
        } else {
          this.staff = this.selectedAppointment.appointment.staffs.find(
            (staff) => staff.id === this.details.appointment_info.staff_id
          );
        }
      }

      if (this.details.appointment_info) {
        if (
          this.details.appointment_info.is_group_appointment &&
          this.details.appointment_info.other_attendees_count > 0 &&
          this.details.appointment_info.meeting_type_id
        ) {
          this.meeting_type_choices_locked = true;
          this.selectedAppointment.meeting_type_id = this.details.appointment_info.meeting_type_id;
        } else {
          this.meeting_type_choices_locked = false;
        }
      }

      setTimeout(() => {
        this.updateFilterDef();
      });
    });

    this.getInstance('availability_meeting_types').addEventListener('output-changed', (availabilityMeetingTypes) => {
      this.meeting_types = availabilityMeetingTypes.detail.value;
    });

    this.getInstance('availability_meeting_types').addEventListener('output-changed', (availabilityMeetingTypes) => {
      this.meeting_types = availabilityMeetingTypes.detail.value;
    });

    this.getInstance('confirm_appointment').addEventListener('output-changed', () => {
      this.isSubmitting = false;
    });
  }

  disconnectedCallback() {
    super.disconnectedCallback();
  }

  static get properties() {
    return {
      /**
       * all links are direct and not altered to work as relative links (client-side)
       */
      isDirect: Boolean,
      /**
       * the navigation bucket
       */
      index: Object
    };
  }

  get defaultScript() {
    return {
      system: {
        state: 'main'
      },
      action: {},
      bucket: {
        context: {},
        record: {}
      },
      io: {}
    };
  }

  getToken() {
    let token;
    const searchParams =
      window.location.hash.split('?').length > 1 ? `?${window.location.hash.split('?')[1]}` : window.location.search;
    const urlParams = new URLSearchParams(searchParams);

    if (urlParams.has('invitation_token')) {
      token = urlParams.get('invitation_token');
    } else if (urlParams.has('token')) {
      token = urlParams.get('token');
    }
    return token;
  }

  render() {
    const allAppointments = this.getDefinition('all_appointments');
    const customContentText = this.getDefinition('custom_content');

    if (!this.selectedAppointment || !allAppointments || !customContentText) {
      return this.__fetchingResultsTemplate();
    } else {
      return html`<ai-screen gid="_screen" ns="${this.ns}">
        <ai-state gid="_state__main" ns="${this.ns}" states="main" class="">
          <div l-kind="hort">
            <div class="select_care_unit_service mbl" l-size="12">
              ${this.__stylesTemplate()} ${this.__groupAppointmentsTemplate()}
              <div class="appointment_details clr-gray_30-bdr bdr-round-all bdr-solid bas pam mbm">
                ${this.__detailsTemplate()}
              </div>
              <div class="form_container clr-gray_30-bdr bdr-round-all bdr-solid bas plm prm">
                ${this.__formTemplate()}
              </div>
            </div>
          </div>
        </ai-state>
      </ai-screen> `;
    }
  }

  getInstance(instanceName) {
    return HIP.db.gid.queryOne(instanceName, this.ns);
  }

  getDefinition(bucketName) {
    const bucket = this.getInstance(bucketName);
    if (bucket) {
      return bucket.output;
    }
  }

  // OPERATING LOGIC

  __submitForm(event) {
    if (this.isSubmitting) return;

    this.isSubmitting = true;
    this.getInstance('submit_record').invoke(event.detail.value);
  }

  updateFilterDef(formData) {
    const def = tt.deepClone(this.getDefinition('record_type') || {});
    const allAppointments = this.getDefinition('all_appointments');
    const customContentText = this.getDefinition('custom_content') || {};

    if (allAppointments && !allAppointments.user) {
      allAppointments.user = {};
    }
    /**
     * Below code is mapping dropdown values & setting picked options from URL.
     */
    if (def && def.properties) {
      if (allAppointments && allAppointments.user && allAppointments.user.cell_phone) {
        def.properties.student_phone.initial_value = allAppointments.user.cell_phone;
      }

      if (this.selectedAppointment) {
        const momentDate = new moment(this.selectedAppointment.begin_time);

        const timeRange = formatRangeTime(
          momentDate,
          momentDate.clone().add(this.selectedAppointment.duration, 'minutes')
        );
        this.date = momentDate.format('MM/DD/YYYY');
        this.time = timeRange;

        if (this.selectedAppointment.meeting_type_id) {
          def.properties.meeting_type_id.initial_value = this.selectedAppointment.meeting_type_id;
        }

        if (this.meeting_types && this.meeting_types.length > 0) {
          def.properties.meeting_type_id.options = this.meeting_types;

          if (this.selectedAppointment.meeting_type_id) {
            def.properties.meeting_type_id.initial_value = this.selectedAppointment.meeting_type_id;
          }
        } else if (!this.selectedAppointment.meeting_types || this.selectedAppointment.meeting_types.length === 0) {
          def.properties.meeting_type_id.hidden = true;
          def.properties.meeting_type_id.disabled = true;
        }

        if (this.meeting_type_choices_locked) {
          def.properties.meeting_type_id.readonly = true;
          def.properties.meeting_type_id.description = 'Cannot change because this is a group appointment.';
        } else if (this.meeting_types) {
          if (this.meeting_types.length === 1) {
            def.properties.meeting_type_id.readonly = true;
            def.properties.meeting_type_id.initial_value = this.meeting_types[0].id;
            def.properties.meeting_type_id.description =
              'You are seeing the only meeting type available for this time slot.';
          } else {
            def.properties.meeting_type_id.description =
              'You are seeing the meeting types available for this time slot.';
          }
        }

        const settings = this.selectedAppointment.communication_settings;
        const email = def.properties.email_reminder;
        const sms = def.properties.sms_reminder;
        sms.readonly = !settings.student_sms_reminder_editable;

        // Form is reloading and we can set the form
        // choices back to their previous values
        if (this.previousFormData) {
          const comments = def.properties.comments;

          comments.initial_value = this.previousFormData.comments;

          email.initial_value = this.previousFormData.email_reminder;
          sms.initial_value = this.previousFormData.sms_reminder;
          sms.readonly = !settings.student_sms_reminder_editable;

          delete this.previousFormData;
        } else {
          email.initial_value = settings.student_email_reminder_default;
          sms.initial_value = settings.student_sms_reminder_default;
        }

        def.properties.student_phone.hidden = !settings.student_sms_reminder_default;
        def.properties.email_reminder.initial_value && allAppointments && allAppointments.user
          ? (def.properties.email_reminder.description = `Reminder will be sent to ${allAppointments.user.email}`)
          : '';
      }

      // Udpate custom labels
      if (customContentText.student_scheduling.care_unit_label) {
        this.care_unit_label = customContentText.student_scheduling.care_unit_label;
      }
      if (customContentText.student_scheduling.service_label) {
        this.service_label = customContentText.student_scheduling.service_label;
      }
      if (customContentText.student_scheduling.location_label) {
        this.location_label = customContentText.student_scheduling.location_label;
      }
      if (customContentText.student_scheduling.staff_label) {
        this.staff_label = customContentText.student_scheduling.staff_label;
      }
      if (customContentText.student_scheduling.course_label) {
        this.course_label = customContentText.student_scheduling.course_label;
      }

      if (formData) {
        def.properties.student_phone.hidden = !formData.sms_reminder;
      }

      // No SMS reminder feature for prospect users at this point
      if (allAppointments.user && allAppointments.user.is_prospect) {
        def.properties.sms_reminder.hidden = true;
        def.properties.student_phone.hidden = true;
      }
      let smsReminderCheckbox = document.querySelector('#_filters_form_builder__sms_reminder');
      if (smsReminderCheckbox) {
        def.properties.student_phone.hidden = !smsReminderCheckbox.checked;
      }
      this.filteredFormDefinition = def;
    }
    this.requestUpdate();
  }

  __updateFormFields(e) {
    const allAppointments = this.getDefinition('all_appointments');
    const formData = e.detail.value.valueAll;
    const props = this.filteredFormDefinition.properties;
    const activeElementId = document.activeElement.id;

    if (allAppointments && !allAppointments.user) {
      allAppointments.user = {};
    }
    // SMS reminder checkbox value has changed
    if (formData && !!props.sms_reminder.initial_value !== !!formData.sms_reminder) {
      props.comments.initial_value = formData.comments;
      props.student_phone.initial_value = formData.student_phone;
      props.sms_reminder.initial_value = !!formData.sms_reminder;
      props.student_phone.hidden = !formData.sms_reminder;
      if (!this.selectedAppointment.communication_settings.student_sms_reminder_editable) {
        props.sms_reminder.readonly = !this.selectedAppointment.communication_settings.student_sms_reminder_editable;
        props.sms_reminder.initial_value = this.selectedAppointment.communication_settings.student_sms_reminder_default;
        props.student_phone.hidden = !this.selectedAppointment.communication_settings.student_sms_reminder_default;
      }
      if (!props.student_phone.hidden && !formData.student_phone) {
        props.student_phone.initial_value = allAppointments.user.cell_phone;
      }
      this.filteredFormDefinition = { ...this.filteredFormDefinition };
      this.requestUpdate();
    }

    // Email reminder checkbox value has changed
    if (props.email_reminder.initial_value !== formData.email_reminder) {
      props.email_reminder.initial_value = !!formData.email_reminder;
      props.email_reminder.description = formData.email_reminder
        ? `Reminder will be sent to ${allAppointments.user.email}`
        : ``;
      props.student_phone.initial_value = formData.student_phone;
      props.comments.initial_value = formData.comments;
      this.filteredFormDefinition = { ...this.filteredFormDefinition };
      this.requestUpdate();
    }

    // Meeting Type value has changed
    if (
      props.meeting_type_id &&
      props.meeting_type_id.inital_value &&
      formData.meeting_type_id &&
      parseInt(props.meeting_type_id.initial_value) !== parseInt(formData.meeting_type_id)
    ) {
      this.selectedAppointment.meeting_type_id = formData.meeting_type_id != null ? formData.meeting_type_id : null;

      if (this.staff) {
        this.selectedAppointment.staff_id = this.staff.id;
      }

      this.isSubmitting = false;
      this.previousFormData = formData;
      this.getDetailsAction(this.selectedAppointment);
    }

    setTimeout(() => {
      // Workaround until we have access to `render_type: boolean_control` in HIL scripts.
      if (document.activeElement.id !== activeElementId) {
        let elementToFocus = HIP.browser.queryOne(activeElementId);
        if (elementToFocus) {
          elementToFocus.focus();
        }
      }
    }, 0);
  }

  // TEMPLATE PARTIALS

  __fetchingResultsTemplate() {
    return html`<div class="text-center pamx min-h-360">
      <div l-kind l-width-fixed l-align="center" l-justify="center" class="min-h-120">
        <hi-indicator></hi-indicator>
      </div>
      <p class="text-sz-lx text-bold">Loading Appointment...</p>
      <p class="clr-gray_70-txt"></p>
    </div>`;
  }

  __stylesTemplate() {
    return html`
      <style>
        dt {
          font-weight: var(--hi-font--weight-normal);
          font-size: var(--hi-font-size--normal);
          padding-bottom: var(--hi-size--small-x);
        }
        dl,
        dd {
          font-size: var(--hi-font-size--normal);
        }
        dl > div {
          border-bottom: 1px solid var(--hi-color--gray-30);
          padding-top: var(--hi-size--medium);
          padding-bottom: var(--hi-size--medium);
        }
        dl > div:last-child {
          border-bottom: none;
        }
        dd {
          color: var(--hi-color--gray-80);
          font-weight: var(--hi-font--weight-light);
        }
        dd > div {
          line-height: 1.5rem;
        }
      </style>
    `;
  }

  __groupAppointmentsTemplate() {
    let group_appointment_text = '';
    if (
      this.staff &&
      this.details &&
      this.details.appointment_info &&
      this.details.appointment_info.is_group_appointment
    ) {
      if (this.details.appointment_info.other_attendees_count === 0) {
        group_appointment_text = `and up to ${this.details.appointment_info.total_possible_attendees - 1} other(s)`;
      } else {
        group_appointment_text = `and ${this.details.appointment_info.other_attendees_count} other(s)`;
      }
    }

    return group_appointment_text
      ? html`<hi-alert theme="warning" class="mbm" mode="presentation"
          ><p>This is a group appointment with ${this.staff.title} ${group_appointment_text}</p></hi-alert
        >`
      : ``;
  }

  __detailsTemplate() {
    const course = this.selectedAppointment.courses.find((c) => {
      return c.id === this.selectedAppointment.course_id;
    });
    return html`
      <dl l-kind="hort" l-justify="center">
        <div l-size="12" l-sml-size="6">
          <dt>${this.care_unit_label}</dt>
          <dt></dt>
          <dd>${this.selectedAppointment.appointment.careUnit.title}</dd>
        </div>
        <div l-size="12" l-sml-size="6">
          <dt>${this.service_label}</dt>
          <dd>${this.selectedAppointment.appointment.service.title}</dd>
        </div>
        <div l-size="12" l-sml-size="6">
          <dt>Date</dt>
          <dd>${this.date}</dd>
        </div>
        <div l-size="12" l-sml-size="6">
          <dt>Time</dt>
          <dd>${this.time}</dd>
        </div>
        <div l-size="12">
          <dt>${this.location_label}</dt>
          <dd>
            <div>${this.selectedAppointment.appointment.location.title}</div>
            <div class="location_description">
              ${this.selectedAppointment.appointment.location.description
                ? html`${unsafeHTML(
                    this.selectedAppointment.appointment.location.description?.replace(
                      /href=/g,
                      'target="_blank" href='
                    )
                  )}`
                : ''}
            </div>
          </dd>
        </div>
        ${this.staff
          ? html`<div l-size="12">
              <dt>${this.staff_label}</dt>
              <dd>${this.staff.title}</dd>
            </div>`
          : ``}
        ${this.details && this.details.availability_info && this.details.availability_info.details
          ? html`<div l-size="12" class="availability_details">
              <dt>Details</dt>
              <dd>${unsafeHTML(this.details.availability_info.details)}</dd>
            </div>`
          : ``}
        ${this.details && this.details.availability_info && this.details.availability_info.meeting_url_or_phone
          ? html`<div l-size="12">
              <dt>URL / Phone Number</dt>
              <dd>${this.details.availability_info.meeting_url_or_phone}</dd>
            </div>`
          : ``}
        ${course
          ? html` <div l-size="12">
              <dt>${this.course_label}</dt>
              <dd>${course.title}</dd>
            </div>`
          : ``}
      </dl>
    `;
  }

  __formTemplate() {
    return html`
      <hi-form
        ns="${this.ns}"
        id="_filters_form"
        gid="_filters_form"
        aesthetic="invisible"
        flush
        has-controls
        disable-http=""
        @evaluate-ended="${this.__updateFormFields}"
        @submit-ended="${this.__submitForm}"
      >
        <ai-builder-form gid="_filters_form_builder" ns="${this.ns}" .input=${this.filteredFormDefinition}>
        </ai-builder-form>
        <hi-button class="mbm" aesthetic="primary" slot="controls" type="submit"
          >${this.student_scheduler_button_text}</hi-button
        >
      </hi-form>
    `;
  }

  // ACTIONS
  getDetailsAction(params = {}) {
    const requestApptAction = this.getInstance('get_details');

    if (this.token) {
      params['token'] = this.token;
    }

    if (this.selectedAppointment.appointment.staffs && this.selectedAppointment.appointment.staffs.length === 1) {
      params['staff_id'] = params.staff_id || this.selectedAppointment.appointment.staffs[0].id;
    }

    if (this.selectedAppointment.meeting_type_id) {
      params['meeting_type_id'] = this.selectedAppointment.meeting_type_id;
    }

    requestApptAction.invoke(params);
  }

  getAllAppointmentsAction(params = {}) {
    const allAppointmentsAction = this.getInstance('get_all_appointments');

    params['slim'] = true;
    if (this.token) {
      params['token'] = this.token;
    }

    allAppointmentsAction.invoke(params);
  }
}

const formatRangeTime = (begin_time, end_time) => {
  const fromTimeFormatted = begin_time.format('h:mm A');
  const toTimeFormatted = end_time.format('h:mm A');
  return `${fromTimeFormatted} - ${toTimeFormatted}`;
};

// customElements.define('ai-prefab-confirm-appointment', PrefabConfirmAppointmentElement);
HIP.factory.element(PrefabConfirmAppointmentElement);
